from typing import *
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.lr_scheduler import StepLR
from torch.optim.lr_scheduler import CyclicLR
from torch.optim.lr_scheduler import ExponentialLR
from torch.optim.lr_scheduler import CosineAnnealingLR
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

from ..misc.toolkit import scheduler_requires_metric

scheduler_dict = {}


def register_scheduler(name: str) -> Callable[[Type], Type]:
    def _register(cls_: Type) -> Type:
        global scheduler_dict
        scheduler_dict[name] = cls_
        return cls_

    return _register


register_scheduler("cyclic")(CyclicLR)
register_scheduler("cosine")(CosineAnnealingLR)
register_scheduler("cosine_restarts")(CosineAnnealingWarmRestarts)


@register_scheduler("step")
class StepLRWithFloor(StepLR):
    def __init__(
        self,
        optimizer: Optimizer,
        step_size: int,
        gamma: float = 0.1,
        last_epoch: int = -1,
        lr_floor: float = 1.0e-8,
    ):
        self.lr_floor = lr_floor
        super().__init__(optimizer, step_size, gamma, last_epoch)

    def get_lr(self) -> List[float]:  # type: ignore
        lrs = super().get_lr()
        return [max(lr, self.lr_floor) for lr in lrs]  # type: ignore

    def _get_closed_form_lr(self) -> List[float]:  # type: ignore
        lrs = super()._get_closed_form_lr()  # type: ignore
        return [max(lr, self.lr_floor) for lr in lrs]


@register_scheduler("exponential")
class ExponentialLRWithFloor(ExponentialLR):
    def __init__(
        self,
        optimizer: Optimizer,
        gamma: float,
        last_epoch: int = -1,
        lr_floor: float = 1.0e-8,
    ):
        self.lr_floor = lr_floor
        super().__init__(optimizer, gamma, last_epoch)

    def get_lr(self) -> List[float]:  # type: ignore
        lrs = super().get_lr()
        return [max(lr, self.lr_floor) for lr in lrs]  # type: ignore

    def _get_closed_form_lr(self) -> List[float]:  # type: ignore
        lrs = super()._get_closed_form_lr()  # type: ignore
        return [max(lr, self.lr_floor) for lr in lrs]


@register_scheduler("plateau")
class ReduceLROnPlateauWithGet(ReduceLROnPlateau):
    def get_lr(self) -> List[float]:
        return [group["lr"] for group in self.optimizer.param_groups]  # type: ignore

    def get_last_lr(self) -> List[float]:
        return self.get_lr()


@register_scheduler("warmup")
class WarmupScheduler(_LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        multiplier: float,
        warmup_step: int,
        scheduler_afterwards_base: Type[_LRScheduler],
        scheduler_afterwards_config: Optional[Dict[str, Any]] = None,
    ):
        self.multiplier = multiplier
        assert self.multiplier > 1.0, "multiplier should be greater than 1"
        self.warmup_step, self.finished_warmup = warmup_step, False
        if scheduler_afterwards_config is None:
            scheduler_afterwards_config = {}
        self.scheduler_afterwards = scheduler_afterwards_base(
            optimizer,
            **scheduler_afterwards_config,
        )
        self.requires_metric = scheduler_requires_metric(self.scheduler_afterwards)
        super().__init__(optimizer)

    @property
    def lr_warmup_func(self) -> Callable[[float], float]:
        multiplier = (self.multiplier - 1.0) * self.last_epoch / self.warmup_step + 1.0  # type: ignore
        return lambda lr: lr * multiplier

    @property
    def lr_multiplier_func(self) -> Callable[[float], float]:
        return lambda lr: lr * self.multiplier

    def get_lr(self) -> List[float]:  # type: ignore
        if self.last_epoch > self.warmup_step:  # type: ignore
            if self.scheduler_afterwards is not None:
                if not self.finished_warmup:
                    self.finished_warmup = True
                    base_lrs = list(
                        map(self.lr_multiplier_func, self.base_lrs)  # type: ignore
                    )
                    self.scheduler_afterwards.base_lrs = base_lrs  # type: ignore
                return self.scheduler_afterwards.get_lr()  # type: ignore
            return list(map(self.lr_multiplier_func, self.base_lrs))
        return list(map(self.lr_warmup_func, self.base_lrs))  # type: ignore

    def get_last_lr(self) -> List[float]:
        if not self.finished_warmup:
            return super().get_last_lr()  # type: ignore
        return self.scheduler_afterwards.get_last_lr()  # type: ignore

    def step(self, metrics: Optional[float] = None) -> None:
        if not self.finished_warmup or self.scheduler_afterwards is None:
            return super().step()
        if not self.requires_metric:
            self.scheduler_afterwards.step()
        else:
            assert metrics is not None
            self.scheduler_afterwards.step(metrics)  # type: ignore


__all__ = ["scheduler_dict", "register_scheduler"]



from .task import Task
from .experiment import inject_distributed_tqdm_kwargs
from .experiment import Experiment
from .experiment import ExperimentResults


__all__ = [
    "Task",
    "inject_distributed_tqdm_kwargs",
    "Experiment",
    "ExperimentResults",
]



import torch

from typing import *
from torch.optim.optimizer import Optimizer

optimizer_dict = {}


def register_optimizer(name: str) -> Callable[[Type], Type]:
    def _register(cls_: Type) -> Type:
        global optimizer_dict
        optimizer_dict[name] = cls_
        return cls_

    return _register


register_optimizer("sgd")(torch.optim.SGD)
register_optimizer("adam")(torch.optim.Adam)
register_optimizer("adamw")(torch.optim.AdamW)
register_optimizer("rmsprop")(torch.optim.RMSprop)


@register_optimizer("nag")
class NAG(Optimizer):
    def __init__(
        self,
        params: Iterable[torch.Tensor],
        lr: float,
        momentum: float = 0.0,
        weight_decay: float = 0.0,
    ):
        defaults = dict(lr=lr, lr_old=lr, momentum=momentum, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure: Optional[Callable] = None) -> Optional[Any]:
        loss = None
        if closure is not None:
            loss = closure()
        for group in self.param_groups:
            weight_decay = group["weight_decay"]
            momentum = group["momentum"]
            lr = group["lr"]
            lr_old = group.get("lr_old", lr)
            lr_correct = lr / lr_old
            for p in group["params"]:
                if p.grad is None:
                    continue
                d_p = p.grad.data
                param_state = self.state[p]
                if "momentum_buffer" not in param_state:
                    param_state["momentum_buffer"] = d_p.clone().zero_()
                buf = param_state["momentum_buffer"]
                if weight_decay != 0:
                    p.data.mul_(1 - lr * weight_decay)
                p.data.add_(buf, alpha=momentum * momentum * lr_correct)
                p.data.add_(d_p, alpha=-(1 + momentum) * lr)
                buf.mul_(momentum * lr_correct).add_(d_p, alpha=-lr)
            group["lr_old"] = lr
        return loss


__all__ = ["optimizer_dict", "register_optimizer"]



import os
import json
import torch
import inspect

from abc import abstractmethod
from abc import ABC
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from argparse import Namespace
from cftool.misc import timestamp
from cftool.misc import update_dict
from cftool.misc import register_core
from cftool.misc import shallow_copy_dict
from cftool.misc import LoggingMixin
from cfdata.tabular import task_type_type
from cfdata.tabular import TimeSeriesConfig

try:
    amp: Optional[Any] = torch.cuda.amp
except:
    amp = None

from .types import general_config_type


configs_dict: Dict[str, Dict[str, Type["Configs"]]] = {}


class Configs(ABC, LoggingMixin):
    def __init__(self, config: Optional[Dict[str, Any]] = None):
        if config is None:
            config = {}
        self.config = config

    @abstractmethod
    def get_default(self) -> Dict[str, Any]:
        pass

    def pop(self) -> Dict[str, Any]:
        default = self.get_default()
        if self.config is None:
            return default
        return update_dict(shallow_copy_dict(self.config), default)

    def setdefault(self, key: str, value: Any) -> Any:
        if self.config is None:
            self.config = {key: value}
            return value
        return self.config.setdefault(key, value)

    @classmethod
    def register(cls, scope: str, name: str) -> Callable[[Type], Type]:
        global configs_dict

        def before(cls_: Type) -> None:
            cls_.name = name

        return register_core(
            name,
            configs_dict.setdefault(scope, {}),
            before_register=before,
        )

    @classmethod
    def get(cls, scope: str, name: str, **kwargs: Any) -> "Configs":
        return configs_dict[scope][name](kwargs)


def _parse_config(config: general_config_type) -> Dict[str, Any]:
    if config is None:
        return {}
    if isinstance(config, str):
        with open(config, "r") as f:
            return json.load(f)
    return shallow_copy_dict(config)


class Elements(NamedTuple):
    model: str = "fcnn"
    ds_args: Optional[Namespace] = None
    aggregator: str = "sum"
    aggregator_config: Optional[Dict[str, Any]] = None
    production: Optional[str] = None
    data_protocol: str = "tabular"
    loader_protocol: str = "tabular"
    sampler_protocol: str = "tabular"
    data_config: Optional[Dict[str, Any]] = None
    task_type: Optional[task_type_type] = None
    use_simplify_data: bool = False
    ts_config: Optional[TimeSeriesConfig] = None
    aggregation: str = "continuous"
    aggregation_config: Optional[Dict[str, Any]] = None
    ts_label_collator_config: Optional[Dict[str, Any]] = None
    delim: Optional[str] = None
    has_column_names: Optional[bool] = None
    read_config: Optional[Dict[str, Any]] = None
    logging_folder: Optional[str] = None
    logging_file: Optional[str] = None
    batch_size: Optional[int] = None
    cv_split: Optional[Union[float, int]] = None
    use_amp: bool = False
    min_epoch: Optional[int] = None
    num_epoch: Optional[int] = None
    max_epoch: Optional[int] = None
    fixed_epoch: Optional[int] = None
    max_snapshot_file: int = 5
    clip_norm: float = 0.0
    ema_decay: float = 0.0
    model_config: Optional[Dict[str, Any]] = None
    loss: str = "auto"
    loss_config: Optional[Dict[str, Any]] = None
    default_encoding_init_method: Optional[str] = None
    metrics: Union[str, List[str]] = "auto"
    metric_config: Optional[Dict[str, Any]] = None
    lr: Optional[float] = None
    optimizer: Optional[str] = "adamw"
    scheduler: Optional[str] = "warmup"
    optimizer_config: Optional[Dict[str, Any]] = None
    scheduler_config: Optional[Dict[str, Any]] = None
    optimizers: Optional[Dict[str, Any]] = None
    verbose_level: int = 2
    use_tqdm: bool = True
    trigger_logging: bool = False
    log_pipeline_to_artifacts: bool = False
    cuda: Optional[Union[int, str]] = None
    mlflow_config: Optional[Dict[str, Any]] = None
    extra_config: Optional[Dict[str, Any]] = None
    user_config: Optional[Dict[str, Any]] = None
    user_increment_config: Optional[Dict[str, Any]] = None

    @staticmethod
    def affected_mappings() -> Dict[str, Set[str]]:
        return {
            "metrics": {"metric_config*"},
            "use_simplify_data": {"simplify"},
            "ts_config": {"time_series_config*"},
            "fixed_epoch": {"min_epoch", "num_epoch", "max_epoch"},
        }

    @property
    def user_defined_config(self) -> Dict[str, Any]:
        user_config = shallow_copy_dict(self.user_config or {})
        user_increment_config = shallow_copy_dict(self.user_increment_config or {})
        return update_dict(user_increment_config, user_config)

    def to_config(self) -> Dict[str, Any]:
        if self.extra_config is None:
            kwargs = {}
        else:
            kwargs = shallow_copy_dict(self.extra_config)
        # inject fields
        spec = inspect.getfullargspec(type(self)).args[1:-2]
        for key in spec:
            value = getattr(self, key)
            if value is not None or key not in kwargs:
                kwargs[key] = value
        # data
        data_config = self.data_config or {}
        task_type = kwargs.pop("task_type")
        if task_type is not None:
            data_config["task_type"] = task_type
        data_config["simplify"] = kwargs.pop("use_simplify_data")
        data_config["time_series_config"] = kwargs.pop("ts_config")
        sampler_config = kwargs.setdefault("sampler_config", {})
        sampler_config["aggregation"] = kwargs.pop("aggregation")
        sampler_config["aggregation_config"] = kwargs.pop("aggregation_config")
        read_config = self.read_config or {}
        read_config["delim"] = kwargs.pop("delim")
        read_config["has_column_names"] = kwargs.pop("has_column_names")
        kwargs["data_config"] = data_config
        kwargs["read_config"] = read_config
        # logging
        if self.logging_folder is not None:
            if self.logging_file is not None:
                logging_file = self.logging_file
            else:
                logging_file = f"{self.model}_{timestamp()}.log"
            kwargs["logging_file"] = logging_file
        # trainer
        trainer_config = kwargs.setdefault("trainer_config", {})
        trainer_config.setdefault("use_amp", kwargs.pop("use_amp"))
        min_epoch = kwargs.pop("min_epoch")
        num_epoch = kwargs.pop("num_epoch")
        max_epoch = kwargs.pop("max_epoch")
        if self.fixed_epoch is not None:
            msg = "`{}` should not be provided when `fixed_epoch` is provided"
            if min_epoch is not None:
                raise ValueError(msg.format("min_epoch"))
            if num_epoch is not None:
                raise ValueError(msg.format("num_epoch"))
            if max_epoch is not None:
                raise ValueError(msg.format("max_epoch"))
            min_epoch = num_epoch = max_epoch = self.fixed_epoch
        no_min, no_num, no_max = min_epoch is None, num_epoch is None, max_epoch is None
        default_min, default_num, default_max = 0, 40, 200
        if no_min and no_num and no_max:
            min_epoch = default_min
            num_epoch = default_num
            max_epoch = default_max
        elif no_min and no_num:
            min_epoch = min(default_min, max_epoch)
            num_epoch = min(default_num, max_epoch)
        elif no_min and no_max:
            min_epoch = min(default_min, num_epoch)
            max_epoch = max(default_max, num_epoch)
        elif no_num and no_max:
            num_epoch = max(default_num, min_epoch)
            max_epoch = max(default_max, min_epoch)
        elif no_min:
            if num_epoch > max_epoch:
                raise ValueError("`num_epoch` should not be greater than `max_epoch`")
            min_epoch = min(default_min, num_epoch)
        elif no_num:
            if min_epoch > max_epoch:
                raise ValueError("`min_epoch` should not be greater than `max_epoch`")
            num_epoch = max(min(default_num, max_epoch), min_epoch)
        elif no_max:
            if min_epoch > num_epoch:
                raise ValueError("`min_epoch` should not be greater than `num_epoch`")
            max_epoch = max(default_max, num_epoch)
        trainer_config.setdefault("min_epoch", min_epoch)
        trainer_config.setdefault("num_epoch", num_epoch)
        trainer_config.setdefault("max_epoch", max_epoch)
        trainer_config.setdefault("max_snapshot_file", kwargs.pop("max_snapshot_file"))
        trainer_config.setdefault("clip_norm", kwargs.pop("clip_norm"))
        # model
        model_config = self.model_config or {}
        model_config["aggregator"] = kwargs.pop("aggregator")
        model_config["aggregator_config"] = kwargs.pop("aggregator_config") or {}
        model_config["ema_decay"] = kwargs.pop("ema_decay")
        model_config["loss"] = kwargs.pop("loss")
        model_config["loss_config"] = kwargs.pop("loss_config") or {}
        default_encoding_init_method = kwargs.pop("default_encoding_init_method")
        if default_encoding_init_method is not None:
            de_cfg = model_config.setdefault("default_encoding_configs", {})
            de_cfg["init_method"] = default_encoding_init_method
        kwargs["model_config"] = model_config
        # metrics
        metric_config = kwargs.pop("metric_config") or {}
        if self.metric_config is None and self.metrics is not None:
            metric_config["types"] = kwargs.pop("metrics")
        metric_config.setdefault("decay", 0.1)
        trainer_config.setdefault("metric_config", metric_config)
        # optimizers
        lr = kwargs.pop("lr")
        optimizer = kwargs.pop("optimizer")
        scheduler = kwargs.pop("scheduler")
        optimizers = kwargs.pop("optimizers")
        optimizer_config = kwargs.pop("optimizer_config") or {}
        scheduler_config = kwargs.pop("scheduler_config") or {}
        if optimizers is not None:
            if optimizer is not None:
                print(
                    f"{LoggingMixin.warning_prefix}`optimizer` is set to "
                    f"'{optimizer}' but `optimizers` is provided, so "
                    "`optimizer` will be ignored"
                )
            if optimizer_config is not None:
                print(
                    f"{LoggingMixin.warning_prefix}`optimizer_config` is "
                    f"set to '{optimizer_config}' but `optimizers` is provided, "
                    "so `optimizer_config` will be ignored"
                )
            if scheduler is not None:
                print(
                    f"{LoggingMixin.warning_prefix}`scheduler` is set to "
                    f"'{scheduler}' but `optimizers` is provided, so "
                    "`scheduler` will be ignored"
                )
            if scheduler_config is not None:
                print(
                    f"{LoggingMixin.warning_prefix}`scheduler_config` is "
                    f"set to '{scheduler_config}' but `optimizers` is provided, "
                    "so `scheduler_config` will be ignored"
                )
        else:
            if lr is not None:
                optimizer_config["lr"] = lr
            optimizers = {
                "all": {
                    "optimizer": optimizer,
                    "scheduler": scheduler,
                    "optimizer_config": optimizer_config,
                    "scheduler_config": scheduler_config,
                }
            }
        if optimizers is not None:
            trainer_config.setdefault("optimizers", optimizers)
        # inject user defined configs
        update_dict(self.user_defined_config, kwargs)
        return kwargs

    @classmethod
    def make(
        cls,
        config: Optional[Dict[str, Any]] = None,
        increment_config: Optional[Dict[str, Any]] = None,
    ) -> "Elements":
        cfg, inc_cfg = map(_parse_config, [config, increment_config])
        kwargs = update_dict(shallow_copy_dict(inc_cfg), shallow_copy_dict(cfg))
        user_cfg, user_inc_cfg = map(shallow_copy_dict, [cfg, inc_cfg])
        # pipeline general
        model = kwargs.setdefault("model", "fcnn")
        kwargs.setdefault("use_binary_threshold", True)
        kwargs.setdefault("data_config", {})
        kwargs.setdefault("read_config", {})
        kwargs.setdefault("cv_split", None)
        kwargs.setdefault("min_cv_split", 100)
        kwargs.setdefault("max_cv_split", 10000)
        kwargs.setdefault("max_cv_split_ratio", 0.5)
        kwargs.setdefault("cv_split_order", "auto")
        kwargs.setdefault("binary_config", {})
        kwargs.setdefault("shuffle_tr", True)
        kwargs.setdefault("cv_batch_size", 512)
        kwargs.setdefault("ts_label_collator_config", {})
        log_folder = kwargs.setdefault("logging_folder", os.path.join("_logs", model))
        log_file = kwargs.get("logging_file")
        if log_file is not None:
            log_path = os.path.join(log_folder, log_file)
        else:
            log_path = os.path.abspath(os.path.join(log_folder, f"{timestamp()}.log"))
        kwargs["_logging_path_"] = log_path
        # trainer general
        trainer_config = kwargs.setdefault("trainer_config", {})
        trainer_config.setdefault("update_binary_threshold_at_runtime", False)
        use_amp = trainer_config.get("use_amp", False)
        trainer_config["use_amp"] = use_amp and amp is not None
        default_checkpoint_folder = os.path.join(log_folder, "checkpoints")
        trainer_config.setdefault("checkpoint_folder", default_checkpoint_folder)
        # model general
        model_config = kwargs.setdefault("model_config", {})
        encoding_methods = model_config.get("encoding_methods", {})
        encoding_configs = model_config.get("encoding_configs", {})
        encoding_methods = {str(k): v for k, v in encoding_methods.items()}
        encoding_configs = {str(k): v for k, v in encoding_configs.items()}
        model_config["encoding_methods"] = encoding_methods
        model_config["encoding_configs"] = encoding_configs
        model_config.setdefault("default_encoding_configs", {})
        model_config.setdefault("loss_config", {})
        # misc
        kwargs.setdefault("cuda", None)
        kwargs.setdefault("trial", None)
        kwargs.setdefault("use_tqdm", True)
        kwargs.setdefault("verbose_level", 2)
        kwargs.setdefault("mlflow_config", None)
        kwargs.setdefault("trigger_logging", False)
        kwargs.setdefault("use_timing_context", True)
        # convert to `Elements`
        spec = inspect.getfullargspec(cls).args[1:-2]
        main_configs = {key: kwargs.pop(key) for key in spec if key in kwargs}
        existing_extra_config = main_configs.get("extra_config")
        if existing_extra_config is None:
            main_configs["extra_config"] = kwargs
        else:
            update_dict(kwargs, existing_extra_config)
        main_configs["user_config"] = user_cfg
        main_configs["user_increment_config"] = user_inc_cfg
        return cls(**main_configs)


class Environment:
    def __init__(
        self,
        config: Dict[str, Any],
        user_config: Optional[Dict[str, Any]] = None,
        user_increment_config: Optional[Dict[str, Any]] = None,
        set_device: bool = True,
    ):
        self.config = config
        self.user_config = user_config or {}
        self.user_increment_config = user_increment_config or {}
        # deep speed
        self.is_rank_0 = True
        if self.deepspeed:
            self.is_rank_0 = self.local_rank is None or self.local_rank == 0
            logging_folder = config.pop("logging_folder")
            current_timestamp = timestamp(ensure_different=True)
            config["logging_folder"] = os.path.join(logging_folder, current_timestamp)
            if set_device:
                if self.local_rank is None:
                    self.ds_args.local_rank = 0
                torch.cuda.set_device(self.local_rank)

    def __getattr__(self, item: str) -> Any:
        return self.config[item]

    @property
    def user_defined_config(self) -> Dict[str, Any]:
        user_cfg = shallow_copy_dict(self.user_config)
        user_inc_cfg = shallow_copy_dict(self.user_increment_config)
        return update_dict(user_inc_cfg, user_cfg)

    def update_default_config(self, new_default_config: Dict[str, Any]) -> None:
        user_affected = set()
        affected_mappings = Elements.affected_mappings()

        def _inject_affected(current: Dict[str, Any]) -> None:
            for key, value in current.items():
                user_affected.update(affected_mappings.get(key, {key}))
                if isinstance(value, dict):
                    _inject_affected(value)

        def _core(current: Dict[str, Any], new_default: Dict[str, Any]) -> None:
            for k, new_default_v in new_default.items():
                current_v = current.get(k)
                if current_v is None:
                    current[k] = new_default_v
                else:
                    if f"{k}*" in user_affected:
                        continue
                    if not isinstance(new_default_v, dict):
                        if k not in user_affected:
                            current[k] = new_default_v
                    else:
                        _core(current_v, new_default_v)

        _inject_affected(self.user_defined_config)
        _core(self.config, new_default_config)

    @property
    def deepspeed(self) -> bool:
        return self.ds_args is not None

    @property
    def local_rank(self) -> int:
        return self.ds_args.local_rank

    @property
    def device(self) -> torch.device:
        if self.deepspeed:
            return torch.device("cuda", self.local_rank)
        cuda = self.cuda
        if cuda == "cpu":
            return torch.device("cpu")
        if cuda is not None:
            return torch.device(f"cuda:{cuda}")
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")

    @property
    def pipeline_config(self) -> Dict[str, Any]:
        return self.config

    @property
    def trainer_config(self) -> Dict[str, Any]:
        return self.config["trainer_config"]

    @property
    def model_config(self) -> Dict[str, Any]:
        return self.config["model_config"]

    @classmethod
    def from_elements(
        cls,
        elements: Elements,
        set_device: bool = True,
    ) -> "Environment":
        return cls(
            elements.to_config(),
            elements.user_config,
            elements.user_increment_config,
            set_device,
        )


__all__ = ["configs_dict", "Configs", "Elements", "Environment"]



import math
import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Callable
from typing import Optional
from torch.nn import Module
from torch.nn import ModuleList
from cftool.misc import context_error_handler

from .components import *
from ..base import HeadBase
from ...blocks import InvertibleBlock
from ...blocks import PseudoInvertibleBlock
from ....types import tensor_dict_type
from ....misc.toolkit import switch_requires_grad


@HeadBase.register("ddr")
class DDRHead(HeadBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        fetch_q: bool,
        fetch_cdf: bool,
        num_layers: int,
        num_blocks: int,
        latent_dim: int,
        **kwargs: Any,
    ):
        super().__init__(in_dim, out_dim, **kwargs)
        assert out_dim == 1
        # common
        self.register_buffer("cdf_logit_anchor", torch.tensor([math.log(3.0)]))
        if not fetch_q and not fetch_cdf:
            raise ValueError("something must be fetched, either `q` or `cdf`")
        self.fetch_q = fetch_q
        self.fetch_cdf = fetch_cdf
        if num_blocks % 2 != 0:
            raise ValueError("`num_blocks` should be divided by 2")
        self.latent_dim = latent_dim
        # median mappings
        median_units = [latent_dim] * num_layers
        self.median_mappings = get_cond_mappings(in_dim, 3, median_units, False)
        # pseudo invertible q
        kwargs = {"num_layers": num_layers, "condition_dim": in_dim}
        if not self.fetch_q:
            q_to_latent_builder = self.dummy_builder
        else:
            q_to_latent_builder = monotonous_builder(
                is_q=True,
                ascent1=True,
                ascent2=True,
                to_latent=True,
                **kwargs,  # type: ignore
            )
        if not self.fetch_cdf:
            q_from_latent_builder = self.dummy_builder
        else:
            q_from_latent_builder = monotonous_builder(
                is_q=True,
                ascent1=True,
                ascent2=True,
                to_latent=False,
                **kwargs,  # type: ignore
            )
        self.q_invertible = PseudoInvertibleBlock(
            1,
            latent_dim,
            3,
            to_transition_builder=q_to_latent_builder,
            from_transition_builder=q_from_latent_builder,
        )
        # pseudo invertible y
        if not self.fetch_cdf:
            y_to_latent_builder = self.dummy_builder
        else:
            y_to_latent_builder = monotonous_builder(
                is_q=False,
                ascent1=True,
                ascent2=num_blocks == 0,
                to_latent=True,
                **kwargs,  # type: ignore
            )
        if not self.fetch_q:
            y_from_latent_builder = self.dummy_builder
        else:
            y_from_latent_builder = monotonous_builder(  # type: ignore
                is_q=False,
                ascent1=True,
                ascent2=True,
                to_latent=False,
                cond_mappings=self.median_mappings,
                **kwargs,
            )
        self.y_invertible = PseudoInvertibleBlock(
            1,
            latent_dim,
            3,
            to_transition_builder=y_to_latent_builder,
            from_transition_builder=y_from_latent_builder,
        )
        # q parameters
        if not self.fetch_q:
            self.q_parameters = []
        else:
            q_params1 = list(self.q_invertible.to_latent.parameters())
            q_params2 = list(self.y_invertible.from_latent.parameters())
            self.q_parameters = q_params1 + q_params2
        # invertible blocks
        self.num_blocks = num_blocks
        self.block_parameters: List[nn.Parameter] = []
        self.blocks = ModuleList()
        for _ in range(num_blocks):
            block = InvertibleBlock(latent_dim, transition_builder=transition_builder)
            self.block_parameters.extend(block.parameters())
            self.blocks.append(block)

    @property
    def q_fn(self) -> Callable[[Tensor], Tensor]:
        return lambda q: 2.0 * q - 1.0

    @property
    def q_inv_fn(self) -> Callable[[Tensor], Tensor]:
        return torch.sigmoid

    @property
    def dummy_builder(self) -> Callable[[int, int], Module]:
        return lambda _, __: nn.Identity()

    def _detach_q(self) -> context_error_handler:
        fetch_q = self.fetch_q

        def switch(requires_grad: bool) -> None:
            switch_requires_grad(self.q_parameters, requires_grad)
            switch_requires_grad(self.block_parameters, requires_grad)

        class _(context_error_handler):
            def __enter__(self) -> None:
                switch(not fetch_q)

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                switch(True)

        return _()

    @staticmethod
    def _get_median_outputs(pack: Pack) -> tensor_dict_type:
        cond_net = pack.cond
        med, pos_med_res, neg_med_res = cond_net.split(1, dim=1)
        return {"median": med, "pos_med_res": pos_med_res, "neg_med_res": neg_med_res}

    def _get_median_pack(self, net: Tensor) -> Pack:
        zeros = torch.zeros_like(net)
        res1, res2 = [], []
        for i, mapping in enumerate(self.median_mappings):
            net = mapping(net)
            if i == len(self.median_mappings) - 1:
                res1.append(zeros)
                res2.append(zeros)
            else:
                chunked = net.chunk(2, dim=1)
                res1.append(chunked[0])
                res2.append(chunked[1])
        return Pack(zeros, net, (res1, res2))

    @staticmethod
    def _merge_y_pack(
        pack: Pack,
        q_batch: Tensor,
        median_outputs: tensor_dict_type,
    ) -> tensor_dict_type:
        pos_med_res = median_outputs["pos_med_res"]
        neg_med_res = median_outputs["neg_med_res"]
        q_positive_mask = torch.sign(q_batch) == 1.0
        med_res = torch.where(q_positive_mask, pos_med_res, neg_med_res).detach()
        y_mul = pack.net
        y_res = med_res * y_mul
        return {
            "y_res": y_res,
            "med_mul": y_mul,
            "med_res": med_res,
            "q_positive_mask": q_positive_mask,
        }

    def _merge_q_pack(self, pack: Pack) -> tensor_dict_type:
        q_logit_mul = pack.net
        q_logit = self.cdf_logit_anchor * q_logit_mul
        return {
            "q": self.q_inv_fn(q_logit),
            "q_logit": q_logit,
            "q_logit_mul": q_logit_mul,
        }

    def _q_results(
        self,
        net: Tensor,
        q_batch: Tensor,
        median_outputs: tensor_dict_type,
        median_responses: responses_tuple_type,
        do_inverse: bool = False,
    ) -> tensor_dict_type:
        # prepare q_latent
        q_batch = self.q_fn(q_batch)
        q1, q2 = self.q_invertible(q_batch, net)
        q1, q2 = q1.net, q2.net
        # simulate quantile function
        for block in self.blocks:
            q1, q2 = block(q1, q2)
        y_pack = self.y_invertible.inverse((q1, q2), median_responses)
        results = self._merge_y_pack(y_pack, q_batch, median_outputs)  # type: ignore
        q_inverse = None
        if do_inverse and self.fetch_cdf:
            y_res = results["y_res"]
            assert y_res is not None
            inverse_results = self._y_results(
                net,
                y_res.detach(),
                median_outputs,
                median_responses,
            )
            q_inverse = inverse_results["q"]
        results["q_inverse"] = q_inverse
        return results

    def _y_results(
        self,
        net: Tensor,
        y_batch: Tensor,
        median_outputs: tensor_dict_type,
        median_responses: responses_tuple_type,
        do_inverse: bool = False,
    ) -> tensor_dict_type:
        # prepare y_latent
        y1, y2 = self.y_invertible(y_batch, net)
        y1, y2 = y1.net, y2.net
        # simulate cdf
        with self._detach_q():
            for i in range(self.num_blocks):
                y1, y2 = self.blocks[self.num_blocks - i - 1].inverse(y1, y2)
            q_logit_pack = self.q_invertible.inverse((y1, y2), median_responses)
            results = self._merge_q_pack(q_logit_pack)  # type: ignore
            y_inverse_res = None
            if do_inverse and self.fetch_q:
                inverse_results = self._q_results(
                    net,
                    results["q"].detach(),
                    median_outputs,
                    median_responses,
                )
                y_inverse_res = inverse_results["y_res"]
            results["y_inverse_res"] = y_inverse_res
        return results

    def forward(
        self,
        net: Tensor,
        *,
        median: bool = False,
        do_inverse: bool = False,
        q_batch: Optional[Tensor] = None,
        y_batch: Optional[Tensor] = None,
    ) -> tensor_dict_type:
        median_pack = self._get_median_pack(net)
        median_responses = median_pack.responses_tuple
        results = median_outputs = self._get_median_outputs(median_pack)
        if self.fetch_q and not median and q_batch is not None:
            results.update(
                self._q_results(
                    net,
                    q_batch,
                    median_outputs,
                    median_responses,
                    do_inverse,
                )
            )
        if self.fetch_cdf and not median and y_batch is not None:
            results.update(
                self._y_results(
                    net,
                    y_batch,
                    median_outputs,
                    median_responses,
                    do_inverse,
                )
            )
        return results


__all__ = ["DDRHead"]



import torch

import torch.nn as nn

from typing import *

from ..base import HeadBase
from ...blocks import DNDF
from ...blocks import TreeResBlock


@HeadBase.register("tree_stack")
class TreeStackHead(HeadBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_blocks: int,
        dndf_config: Dict[str, Any],
        out_dndf_config: Dict[str, Any],
        **kwargs: Any,
    ):
        super().__init__(in_dim, out_dim, **kwargs)
        self.res_blocks = nn.ModuleList()
        for _ in range(num_blocks):
            self.res_blocks.append(TreeResBlock(in_dim, dndf_config))
        self.out_dndf = DNDF(
            in_dim,
            out_dim,
            **out_dndf_config,
        )

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        for block in self.res_blocks:
            net = block(net)
        return self.out_dndf(net)


__all__ = ["TreeStackHead"]



import numpy as np

from typing import *
from cftool.ml import EnsemblePattern

from ..protocol import DataProtocol


class TSLabelCollator:
    custom_methods: Dict[str, Callable[[np.ndarray], np.ndarray]] = {}

    def __init__(
        self,
        data: DataProtocol,
        config: Optional[Dict[str, Any]] = None,
    ):
        self.data = data
        if config is None:
            config = {}
        self._init_config(config)

    def __call__(self, labels: np.ndarray) -> np.ndarray:
        return self.fn(labels)

    def _init_config(self, config: Dict[str, Any]) -> None:
        self.config = config
        self._method = config.setdefault("method", "average")
        if self._method == "average":
            config.setdefault("num_history", 1)

    @property
    def fn(self) -> Callable[[np.ndarray], np.ndarray]:
        custom_method = TSLabelCollator.custom_methods.get(self._method)
        if custom_method is not None:
            return custom_method
        return getattr(self, f"_{self._method}")

    @staticmethod
    def _last(labels: np.ndarray) -> np.ndarray:
        return labels[..., -1, :]

    def _average(self, labels: np.ndarray) -> np.ndarray:
        num_history = self.config["num_history"]
        extracted = labels[..., -num_history:, :]
        if self.data.is_reg:
            return extracted.mean(axis=1)
        return EnsemblePattern.vote(extracted.squeeze(2), self.data.num_classes)


__all__ = ["TSLabelCollator"]



import math
import torch
import logging

import numpy as np
import torch.nn as nn

from typing import *
from abc import ABCMeta
from collections import defaultdict
from cfdata.tabular.misc import np_int_type

from ..protocol import DataLoaderProtocol
from ..misc.toolkit import to_torch
from ..misc.toolkit import Lambda
from ..misc.toolkit import Initializer
from ..misc.toolkit import LoggingMixinWithRank
from ..modules.auxiliary import Dropout


class EncodingResult(NamedTuple):
    one_hot: Optional[torch.Tensor]
    embedding: Optional[torch.Tensor]

    @property
    def merged(self) -> torch.Tensor:
        if self.one_hot is None and self.embedding is None:
            raise ValueError("no data is provided in `EncodingResult`")
        if self.one_hot is None:
            assert self.embedding is not None
            return self.embedding
        if self.embedding is None:
            assert self.one_hot is not None
            return self.one_hot
        return torch.cat([self.one_hot, self.embedding], dim=1)


class OneHot(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        one_hot_fn = lambda column: nn.functional.one_hot(column, dim)
        self.core = Lambda(one_hot_fn, f"one_hot_{dim}")
        self.dim = dim

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.core(tensor).to(torch.float32)


class Embedding(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        init_method: Optional[str],
        init_config: Dict[str, Any],
    ):
        super().__init__()
        weights = torch.empty(in_dim, out_dim)
        if init_method is None:
            nn.init.normal_(weights)
        else:
            Initializer(init_config).initialize(weights, init_method)
        self.weights = nn.Parameter(weights)
        embedding_fn = lambda column: nn.functional.embedding(column, self.weights)
        self.core = Lambda(embedding_fn, f"embedding: {in_dim} -> {out_dim}")
        self.in_dim, self.out_dim = in_dim, out_dim

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.core(tensor)


class Encoder(nn.Module, LoggingMixinWithRank, metaclass=ABCMeta):
    def __init__(
        self,
        config: Dict[str, Any],
        input_dims: List[int],
        methods_list: List[Union[str, List[str]]],
        configs: List[Dict[str, Any]],
        categorical_columns: List[int],
        loaders: Dict[str, DataLoaderProtocol],
    ):
        super().__init__()
        self._fe_init_method: Optional[str]
        self._fe_init_config: Optional[Dict[str, Any]]
        self._init_config(config)
        for loader in loaders.values():
            if loader.enabled_sampling:
                raise ValueError("`loader` should not enable sampling in `Encoder`")
            if loader.sampler.shuffle:
                raise ValueError("`loader` should not be shuffled in `Encoder`")
        self.merged_dim = 0
        self.one_hot_dim = 0
        self.embedding_dim = 0
        dims_tensor = torch.tensor(input_dims, dtype=torch.float32)
        self.register_buffer("input_dims", dims_tensor)
        self.tgt_columns = np.array(sorted(categorical_columns), np_int_type)
        self.merged_dims: Dict[int, int] = defaultdict(int)
        self.embeddings = nn.ModuleList()
        self.one_hot_encoders = nn.ModuleList()
        self._one_hot_indices: List[int] = []
        self._embed_indices: List[int] = []
        self._embed_dims: List[int] = []
        for i, (in_dim, methods, config) in enumerate(
            zip(input_dims, methods_list, configs)
        ):
            if isinstance(methods, str):
                methods = [methods]
            self._register(i, in_dim, methods, config)
        # fast embedding
        if self.use_embedding and self._use_fast_embed:
            if isinstance(self._unified_embed_dim, int):
                unified_embed_dim = self._unified_embed_dim
            else:
                if self._unified_embed_dim == "max":
                    unified_embed_dim = max(self._embed_dims)
                elif self._unified_embed_dim == "mean":
                    unified_embed_dim = int(
                        round(sum(self._embed_dims) / self.num_embedding)
                    )
                elif self._unified_embed_dim == "median":
                    half_idx = self.num_embedding // 2
                    sorted_dims = sorted(self._embed_dims)
                    if self.num_embedding % 2 != 0:
                        unified_embed_dim = sorted_dims[half_idx]
                    else:
                        left = sorted_dims[half_idx - 1]
                        right = sorted_dims[half_idx]
                        unified_embed_dim = int(round(0.5 * (left + right)))
                else:
                    raise NotImplementedError(
                        f"unified embedding dim '{self._unified_embed_dim}' "
                        "is not recognized"
                    )
            for i in self._embed_indices:
                self.merged_dims[i] += unified_embed_dim
            embedding_dim_sum = unified_embed_dim * self.num_embedding
            self.embedding_dim = embedding_dim_sum
            self.merged_dim += embedding_dim_sum
            if self._fe_init_method is None:
                self._fe_init_method = self._de_init_method
            if self._fe_init_config is None:
                self._fe_init_config = self._de_init_config
            assert isinstance(self._fe_init_config, dict)
            self.embeddings.append(
                Embedding(
                    sum(input_dims),
                    unified_embed_dim,
                    self._fe_init_method,
                    self._fe_init_config,
                )
            )
            assert isinstance(self.input_dims, torch.Tensor)
            embed_dims_cumsum = self.input_dims[self._embed_indices].cumsum(0)[:-1]
            self.register_buffer("embed_dims_cumsum", embed_dims_cumsum)
        # embedding dropout
        self.embedding_dropout = None
        if self.use_embedding and 0.0 < self._embed_drop < 1.0:
            self.embedding_dropout = Dropout(self._embed_drop)
        # compile
        self.one_hot_columns = self.tgt_columns[self._one_hot_indices]
        self.embedding_columns = self.tgt_columns[self._embed_indices]
        self._all_one_hot = len(self.one_hot_columns) == len(input_dims)
        self._all_embedding = len(self.embedding_columns) == len(input_dims)
        self._compile(loaders)

    @property
    def num_one_hot(self) -> int:
        return len(self._one_hot_indices)

    @property
    def num_embedding(self) -> int:
        return len(self._embed_indices)

    @property
    def use_one_hot(self) -> bool:
        return self.num_one_hot > 0

    @property
    def use_embedding(self) -> bool:
        return self.num_embedding > 0

    def forward(
        self,
        x_batch: torch.Tensor,
        batch_indices: Optional[np.ndarray],
        loader_name: Optional[str],
    ) -> EncodingResult:
        keys = None
        if loader_name is not None:
            keys = self._get_cache_keys(loader_name)
        categorical_columns = x_batch[..., self.tgt_columns]
        if batch_indices is None or loader_name is None:
            self._oob_imputation(categorical_columns)
        use_cache = keys is not None and batch_indices is not None
        # one hot
        if not self.use_one_hot:
            one_hot = None
        else:
            if use_cache:
                one_hot = getattr(self, keys["one_hot"])[batch_indices]  # type: ignore
            else:
                one_hot_columns = categorical_columns
                if not self._all_one_hot:
                    one_hot_columns = one_hot_columns[..., self._one_hot_indices]
                one_hot = self._one_hot(one_hot_columns)
        # embedding
        if not self.use_embedding:
            embedding = None
        else:
            if not use_cache:
                indices = categorical_columns
            else:
                indices = getattr(self, keys["indices"])[batch_indices]  # type: ignore
            if not self._all_embedding:
                indices = indices[..., self._embed_indices]
            if not use_cache and self._use_fast_embed:
                indices[..., 1:] += self.embed_dims_cumsum
                indices = indices.to(torch.long)
            embedding = self._embedding(indices)
            if self.embedding_dropout is not None:
                embedding = self.embedding_dropout(embedding)
        return EncodingResult(one_hot, embedding)

    def _init_config(self, config: Dict[str, Any]) -> None:
        self.config = config
        self._embed_drop = config.setdefault("embedding_dropout", 0.2)
        self._de_init_method = config.setdefault(
            "default_embedding_init_method", "truncated_normal"
        )
        self._de_init_config = config.setdefault(
            "default_embedding_init_config", {"mean": 0.0, "std": 0.02}
        )
        self._use_fast_embed = config.setdefault("use_fast_embedding", True)
        # [ mean | median | max | int ]
        self._unified_embed_dim = config.setdefault("unified_embedding_dim", "max")
        self._fe_init_method = config.setdefault("fast_embedding_init_method", None)
        self._fe_init_config = config.setdefault("fast_embedding_init_config", None)

    def _register(
        self,
        i: int,
        in_dim: int,
        methods: List[str],
        config: Dict[str, Any],
    ) -> None:
        for method in methods:
            attr = getattr(self, f"_register_{method}", None)
            if attr is None:
                msg = f"encoding method '{method}' is not implemented"
                raise NotImplementedError(msg)
            attr(i, in_dim, config)

    def _register_one_hot(self, i: int, in_dim: int, _: Dict[str, Any]) -> None:
        self.one_hot_encoders.append(OneHot(in_dim))
        self._one_hot_indices.append(i)
        self.merged_dims[i] += in_dim
        self.one_hot_dim += in_dim
        self.merged_dim += in_dim

    @staticmethod
    def _get_embed_key(num: int) -> str:
        return f"embedding_weight_{num}"

    def _register_embedding(self, i: int, in_dim: int, config: Dict[str, Any]) -> None:
        self._embed_indices.append(i)
        embedding_dim = config.setdefault("embedding_dim", "auto")
        if isinstance(embedding_dim, int):
            out_dim = embedding_dim
        elif embedding_dim == "log":
            out_dim = math.ceil(math.log2(in_dim))
        elif embedding_dim == "sqrt":
            out_dim = math.ceil(math.sqrt(in_dim))
        elif embedding_dim == "auto":
            out_dim = min(in_dim, max(4, min(8, math.ceil(math.log2(in_dim)))))
        else:
            raise ValueError(f"embedding dim '{embedding_dim}' is not defined")
        if self._use_fast_embed:
            self._embed_dims.append(out_dim)
            if self._fe_init_method is None:
                self._fe_init_method = config.get("init_method")
            if self._fe_init_config is None:
                self._fe_init_config = config.get("init_config")
        else:
            init_method = config.setdefault("init_method", self._de_init_method)
            init_config = config.setdefault("init_config", self._de_init_config)
            self.merged_dims[i] += out_dim
            self.embedding_dim += out_dim
            self.merged_dim += out_dim
            self.embeddings.append(Embedding(in_dim, out_dim, init_method, init_config))

    def _oob_imputation(
        self,
        categorical_columns: torch.Tensor,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
    ) -> None:
        if loader_name is None or batch_indices is None:
            oob_mask = categorical_columns >= self.input_dims
        else:
            keys = self._get_cache_keys(loader_name)
            oob_mask = getattr(self, keys["oob"])[batch_indices]
        if torch.any(oob_mask):
            self.log_msg(  # type: ignore
                "out of bound occurred, "
                f"ratio : {torch.mean(oob_mask.to(torch.float32)).item():8.6f}",
                prefix=self.warning_prefix,  # type: ignore
                verbose_level=5,
                msg_level=logging.WARNING,
            )
            # TODO : currently pytorch does not support onnx with bool masks
            #        in the future this line should be un-indented
            categorical_columns[oob_mask] = 0.0

    @staticmethod
    def _to_split(columns: torch.Tensor) -> List[torch.Tensor]:
        return list(columns.to(torch.long).t().unbind())

    def _one_hot(self, one_hot_columns: torch.Tensor) -> torch.Tensor:
        split = self._to_split(one_hot_columns)
        encodings = [
            encoder(flat_feature)
            for encoder, flat_feature in zip(self.one_hot_encoders, split)
        ]
        return torch.cat(encodings, dim=1)

    def _embedding(self, indices_columns: torch.Tensor) -> torch.Tensor:
        if self._use_fast_embed:
            embed_mat = self.embeddings[0](indices_columns)
            return embed_mat.view(-1, self.embedding_dim)
        split = self._to_split(indices_columns)
        encodings = [
            embedding(flat_feature)
            for embedding, flat_feature in zip(self.embeddings, split)
        ]
        return torch.cat(encodings, dim=1)

    @staticmethod
    def _get_cache_keys(name: str) -> Dict[str, str]:
        return {
            "one_hot": f"{name}_one_hot_cache",
            "indices": f"{name}_indices_cache",
            "oob": f"{name}_oob_cache",
        }

    def _compile(self, loaders: Dict[str, DataLoaderProtocol]) -> None:
        for name, loader in loaders.items():
            categorical_features = []
            return_indices = loader.return_indices
            for sample in loader:
                if return_indices:
                    assert isinstance(sample, tuple)
                    sample = sample[0]
                assert isinstance(sample, dict)
                x_batch = sample["x_batch"]
                categorical_features.append(x_batch[..., self.tgt_columns])
            tensor = to_torch(np.vstack(categorical_features))
            keys = self._get_cache_keys(name)
            # compile oob
            self.register_buffer(keys["oob"], tensor >= self.input_dims)
            self._oob_imputation(
                tensor,
                batch_indices=np.arange(len(tensor)),
                loader_name=name,
            )
            # compile one hot
            if self.use_one_hot:
                one_hot_cache = self._one_hot(tensor[..., self._one_hot_indices])
                self.register_buffer(keys["one_hot"], one_hot_cache)
            # compile embedding
            if self.use_embedding and self._use_fast_embed:
                tensor[..., 1:] += self.embed_dims_cumsum
                indices = tensor.to(torch.long)
                self.register_buffer(keys["indices"], indices)


__all__ = ["Encoder", "EncodingResult"]



import logging

from abc import *
from typing import *
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cftool.misc import LoggingMixin

from .basic import make
from ..pipeline import Pipeline

registered_benchmarks: Dict[str, Dict[str, Dict[str, Any]]] = {}


class Zoo(LoggingMixin, metaclass=ABCMeta):
    token = "_zoo_"

    def __init__(
        self,
        model: str,
        *,
        model_type: str = "default",
        increment_config: Optional[Dict[str, Any]] = None,
    ):
        self.model = model
        self.model_type = model_type
        self.increment_config = increment_config

    @property
    def benchmarks(self) -> Dict[str, Dict[str, Any]]:
        """
        this method should return a dict of configs (which represent benchmarks)
        * Note that "default" key should always be included in the returned dict
        """
        return registered_benchmarks[self.model]

    @property
    def config(self) -> dict:
        """
        return corresponding config of `self.model_type`
        * and update with `increment_config` if provided
        """
        benchmarks = self.benchmarks
        assert "default" in benchmarks, "'default' should be included in config_dict"
        config = benchmarks.get(self.model_type)
        if config is None:
            if self.model_type != "default":
                self.log_msg(
                    f"model_type '{self.model_type}' is not recognized, "
                    "'default' model_type will be used",
                    self.warning_prefix,
                    2,
                    msg_level=logging.WARNING,
                )
                self.model_type = "default"
            config = self.benchmarks["default"]
        new_config = shallow_copy_dict(config)
        if self.increment_config is not None:
            update_dict(self.increment_config, new_config)
        return new_config

    @property
    def m(self) -> Pipeline:
        """ return corresponding model of self.config """
        return make(self.model, **self.config)

    @classmethod
    def parse_type(cls, model_type: str) -> str:
        return f"{cls.token}{model_type}"

    @classmethod
    def register(
        cls,
        model: str,
        model_type: str,
        *,
        transform_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        extractor_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        head_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        increment_configs: Optional[Dict[str, Any]] = None,
    ) -> None:
        global registered_benchmarks
        model_dict = registered_benchmarks.setdefault(model, {})
        pipe_configs: Dict[str, Any] = {}
        if transform_configs is not None:
            for pipe, transform_config in transform_configs.items():
                pipe_configs.setdefault(pipe, {})["transform"] = transform_config
        if extractor_configs is not None:
            for pipe, extractor_config in extractor_configs.items():
                pipe_configs.setdefault(pipe, {})["extractor"] = extractor_config
        if head_configs is not None:
            for pipe, head_config in head_configs.items():
                pipe_configs.setdefault(pipe, {})["head"] = head_config
        config = {}
        if pipe_configs:
            config = {"model_config": {"pipe_configs": pipe_configs}}
        if increment_configs is not None:
            update_dict(increment_configs, config)
        model_dict[model_type] = config


# fcnn

Zoo.register("fcnn", "default")
Zoo.register("fcnn", "light_bn", head_configs={"fcnn": {"hidden_units": [128]}})
Zoo.register(
    "fcnn",
    "on_large",
    head_configs={"fcnn": {"mapping_configs": {"dropout": 0.1, "batch_norm": False}}},
)
Zoo.register(
    "fcnn",
    "light",
    head_configs={
        "fcnn": {"hidden_units": [128], "mapping_configs": {"batch_norm": False}}
    },
)
Zoo.register(
    "fcnn",
    "on_sparse",
    head_configs={
        "fcnn": {
            "hidden_units": [128],
            "mapping_configs": {"dropout": 0.9, "batch_norm": False},
        }
    },
    increment_configs={"optimizer_config": {"lr": 1e-4}},
)

# tree dnn

Zoo.register("tree_dnn", "default")
Zoo.register(
    "tree_dnn",
    "on_large",
    head_configs={
        "dndf": {"dndf_config": None},
        "fcnn": {"mapping_configs": {"dropout": 0.1}},
    },
)
Zoo.register(
    "tree_dnn",
    "light",
    head_configs={
        "dndf": {"dndf_config": None},
        "fcnn": {"mapping_configs": {"batch_norm": False}},
    },
    increment_configs={
        "model_config": {"default_encoding_configs": {"embedding_dim": 8}}
    },
)
Zoo.register(
    "tree_dnn",
    "on_sparse",
    head_configs={
        "dndf": {"dndf_config": None},
        "fcnn": {
            "mapping_configs": {
                "dropout": 0.9,
                "batch_norm": False,
                "pruner_config": None,
            }
        },
    },
    increment_configs={
        "optimizer_config": {"lr": 1e-4},
        "model_config": {"default_encoding_configs": {"embedding_dim": 8}},
    },
)


__all__ = ["Zoo"]



from typing import Any
from typing import Dict

from ..base import HeadConfigs


@HeadConfigs.register("tree_stack", "default")
class DefaultTreeStackConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        return {"num_blocks": 3, "dndf_config": {}, "out_dndf_config": {}}


@HeadConfigs.register("tree_stack", "linear")
class LinearTreeStackConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        return {"num_blocks": 0, "dndf_config": {}, "out_dndf_config": {}}


__all__ = [
    "DefaultTreeStackConfig",
    "LinearTreeStackConfig",
]



import os
import time
import shutil

from typing import Set
from typing import Dict
from typing import List
from typing import Optional
from cftool.misc import LoggingMixin

SAVING_DELIM = "^_^"


def _to_saving_path(identifier: str, saving_folder: Optional[str]) -> str:
    if saving_folder is None:
        saving_path = identifier
    else:
        saving_path = os.path.join(saving_folder, identifier)
    return saving_path


def _make_saving_path(
    i: int,
    name: str,
    saving_path: str,
    remove_existing: bool,
) -> str:
    saving_path = os.path.abspath(saving_path)
    saving_folder, identifier = os.path.split(saving_path)
    postfix = f"{SAVING_DELIM}{name}{SAVING_DELIM}{i:04d}"
    if os.path.isdir(saving_folder) and remove_existing:
        for existing_model in os.listdir(saving_folder):
            if os.path.isdir(os.path.join(saving_folder, existing_model)):
                continue
            if existing_model.startswith(f"{identifier}{postfix}"):
                print(
                    f"{LoggingMixin.warning_prefix}"
                    f"'{existing_model}' was found, it will be removed"
                )
                os.remove(os.path.join(saving_folder, existing_model))
    return f"{saving_path}{postfix}"


def _fetch_saving_paths(
    identifier: str = "cflearn",
    saving_folder: Optional[str] = None,
) -> Dict[str, List[str]]:
    paths: Dict[str, Set[str]] = {}
    saving_path = _to_saving_path(identifier, saving_folder)
    saving_path = os.path.abspath(saving_path)
    base_folder = os.path.dirname(saving_path)
    for existing_model in os.listdir(base_folder):
        if not existing_model.startswith(f"{identifier}{SAVING_DELIM}"):
            continue
        current = os.path.join(base_folder, existing_model)
        if os.path.isdir(current):
            new_path = current
            *folder, name, i = existing_model.split(SAVING_DELIM)
        else:
            existing_model, existing_extension = os.path.splitext(existing_model)
            *folder, name, i = existing_model.split(SAVING_DELIM)
            if existing_extension != ".zip":
                continue
            if os.path.join(base_folder, SAVING_DELIM.join(folder)) != saving_path:
                continue
            new_path = _make_saving_path(int(i), name, saving_path, False)
        paths.setdefault(name, set()).add(new_path)
    return {k: sorted(v) for k, v in paths.items()}


def _remove(identifier: str = "cflearn", saving_folder: str = None) -> None:
    for path_list in _fetch_saving_paths(identifier, saving_folder).values():
        for path in path_list:
            path = f"{path}.zip"
            print(f"{LoggingMixin.info_prefix}removing {path}...")
            os.remove(path)


def _rmtree(folder: str, patience: float = 10.0) -> None:
    if not os.path.isdir(folder):
        return None
    t = time.time()
    while True:
        try:
            if time.time() - t >= patience:
                prefix = LoggingMixin.warning_prefix
                print(f"\n{prefix}failed to rmtree: {folder}")
                break
            shutil.rmtree(folder)
            break
        except:
            print("", end=".", flush=True)
            time.sleep(1)


__all__ = [
    "_to_saving_path",
    "_make_saving_path",
    "_fetch_saving_paths",
    "_remove",
    "_rmtree",
]



from typing import Any

from .base import AggregatorBase
from ...types import tensor_dict_type


@AggregatorBase.register("sum")
class Sum(AggregatorBase):
    def reduce(self, outputs: tensor_dict_type, **kwargs: Any) -> tensor_dict_type:
        values = list(outputs.values())
        output = None
        for value in values:
            if value is None:
                continue
            if output is None:
                output = value
            else:
                output = output + value
        return {"predictions": output}


__all__ = ["Sum"]



from typing import Any
from typing import Dict

from ...configs import Configs


transform_config_mapping = {
    "default": {"one_hot": True, "embedding": True, "only_categorical": False},
    "categorical_only": {"one_hot": True, "embedding": True, "only_categorical": True},
    "one_hot": {"one_hot": True, "embedding": False, "only_categorical": False},
    "one_hot_only": {"one_hot": True, "embedding": False, "only_categorical": True},
    "embedding": {"one_hot": False, "embedding": True, "only_categorical": False},
    "embedding_only": {"one_hot": False, "embedding": True, "only_categorical": True},
    "numerical": {"one_hot": False, "embedding": False, "only_categorical": False},
}


@Configs.register("transform", "default")
class DefaultTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["default"]


@Configs.register("transform", "categorical_only")
class CategoricalOnlyTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["categorical_only"]


@Configs.register("transform", "one_hot")
class OneHotTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["one_hot"]


@Configs.register("transform", "one_hot_only")
class OneHotOnlyTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["one_hot_only"]


@Configs.register("transform", "embedding")
class EmbeddingTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["embedding"]


@Configs.register("transform", "embedding_only")
class EmbeddingOnlyTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["embedding_only"]


@Configs.register("transform", "numerical")
class NumericalTransformConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return transform_config_mapping["numerical"]


__all__ = [
    "transform_config_mapping",
    "DefaultTransformConfig",
    "CategoricalOnlyTransformConfig",
    "OneHotTransformConfig",
    "OneHotOnlyTransformConfig",
    "EmbeddingTransformConfig",
    "EmbeddingOnlyTransformConfig",
    "NumericalTransformConfig",
]



import torch

import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Callable
from cftool.misc import register_core

from ..transform.core import Dimensions


extractor_dict: Dict[str, Type["ExtractorBase"]] = {}


class ExtractorBase(nn.Module, metaclass=ABCMeta):
    def __init__(self, in_flat_dim: int, dimensions: Dimensions, **kwargs: Any):
        super().__init__()
        self.in_flat_dim = in_flat_dim
        self.dimensions = dimensions

    @property
    def in_dim(self) -> int:
        return self.in_flat_dim // self.dimensions.num_history

    @property
    def flatten_ts(self) -> bool:
        return True

    @property
    @abstractmethod
    def out_dim(self) -> int:
        pass

    @abstractmethod
    def forward(self, net: torch.Tensor) -> torch.Tensor:
        pass

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global extractor_dict
        return register_core(name, extractor_dict)

    @classmethod
    def make(
        cls,
        name: str,
        in_dim: int,
        dimensions: Dimensions,
        config: Dict[str, Any],
    ) -> "ExtractorBase":
        return extractor_dict[name](in_dim, dimensions, **config)


__all__ = ["ExtractorBase"]



import torch

from typing import Any
from typing import Dict
from typing import Optional

from ..base import HeadBase
from ...blocks import Linear as Lin


@HeadBase.register("linear")
class LinearHead(HeadBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        linear_config: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(in_dim, out_dim, **kwargs)
        if linear_config is None:
            linear_config = {}
        self.linear = Lin(in_dim, out_dim, **linear_config)

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        return self.linear(net)


__all__ = ["LinearHead"]



import os
import json
import torch
import optuna

import numpy as np
import optuna.visualization as vis

from typing import *
from functools import partial
from tqdm.autonotebook import tqdm
from cftool.misc import shallow_copy_dict
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.ml.utils import scoring_fn_type
from cfdata.tabular import task_type_type
from cfdata.tabular import parse_task_type
from optuna.trial import TrialState
from optuna.trial import FrozenTrial
from optuna.importance import BaseImportanceEvaluator
from plotly.graph_objects import Figure

from .basic import *
from .ensemble import *
from .hpo import optuna_tune
from .hpo import default_scoring
from .hpo import optuna_params_type
from .hpo import OptunaPresetParams
from .production import Pack
from .production import Predictor
from ..types import data_type
from ..types import general_config_type
from ..configs import _parse_config
from ..protocol import DataProtocol


class UnPacked(NamedTuple):
    pattern: EnsemblePattern
    predictors: Dict[str, List[Predictor]]


class Auto:
    data_folder = "__data__"
    data_protocol_file = "data_protocol.txt"
    weights_mapping_file = "weights_mapping.json"

    def __init__(
        self,
        task_type: task_type_type,
        *,
        models: Union[str, List[str]] = "auto",
        tune_lr: bool = True,
        tune_optimizer: bool = True,
        tune_ema_decay: bool = True,
        tune_clip_norm: bool = True,
        tune_init_method: bool = True,
        **kwargs: Any,
    ):
        self.task_type = task_type
        self.preset_params = OptunaPresetParams(
            tune_lr=tune_lr,
            tune_optimizer=tune_optimizer,
            tune_ema_decay=tune_ema_decay,
            tune_clip_norm=tune_clip_norm,
            tune_init_method=tune_init_method,
            **shallow_copy_dict(kwargs),
        )
        # models
        if isinstance(models, list):
            self.models = models
        else:
            if models != "auto":
                self.models = [models]
            else:
                self.models = ["linear", "tree_dnn"]
                parsed_task_type = parse_task_type(task_type)
                if parsed_task_type.is_ts:
                    self.models += ["rnn"]
                elif parsed_task_type.is_clf:
                    self.models += ["nnb", "ndt"]
        if not self.models:
            raise ValueError("`models` should be provided")

    def __str__(self) -> str:
        model_str = ", ".join(self.models)
        return f"Auto({model_str})({self.task_type})"

    __repr__ = __str__

    @property
    def studies(self) -> Dict[str, optuna.study.Study]:
        return {k: v.study for k, v in self.optuna_results.items()}

    @property
    def predict(self) -> Callable:
        return self.pattern.predict

    @property
    def predict_prob(self) -> Callable:
        return partial(self.pattern.predict, requires_prob=True)

    @property
    def pruned_trials(self) -> Dict[str, List[FrozenTrial]]:
        return {
            k: [t for t in v.trials if t.state == TrialState.PRUNED]
            for k, v in self.studies.items()
        }

    @property
    def complete_trials(self) -> Dict[str, List[FrozenTrial]]:
        return {
            k: [t for t in v.trials if t.state == TrialState.COMPLETE]
            for k, v in self.studies.items()
        }

    # api

    def fit(
        self,
        x: data_type,
        y: data_type = None,
        x_cv: data_type = None,
        y_cv: data_type = None,
        *,
        study_config: Optional[Dict[str, Any]] = None,
        predict_config: Optional[Dict[str, Any]] = None,
        metrics: Optional[Union[str, List[str]]] = None,
        params: Optional[Dict[str, optuna_params_type]] = None,
        num_jobs: int = 1,
        num_trial: int = 20,
        num_repeat: int = 5,
        num_parallel: int = 1,
        timeout: Optional[float] = None,
        score_weights: Optional[Dict[str, float]] = None,
        estimator_scoring_function: Union[str, scoring_fn_type] = default_scoring,
        temp_folder: str = "__tmp__",
        num_final_repeat: int = 20,
        extra_config: general_config_type = None,
        cuda: Optional[Union[str, int]] = None,
    ) -> "Auto":
        self.best_params = {}
        self.optuna_results = {}
        extra_config = _parse_config(extra_config)

        tuner = None
        optuna_folder = os.path.join(temp_folder, "__optuna__")
        for model in self.models:
            if params is not None:
                model_params = params[model]
            else:
                model_params = self.preset_params.get(model)

            args = () if tuner is not None else (x, y, x_cv, y_cv)
            optuna_result = optuna_tune(
                *args,
                model=model,
                task_type=self.task_type,
                tuner=tuner,
                params=model_params,
                study_config=study_config,
                metrics=metrics,
                num_jobs=num_jobs,
                num_trial=num_trial,
                num_repeat=num_repeat,
                num_parallel=num_parallel,
                timeout=timeout,
                score_weights=score_weights,
                estimator_scoring_function=estimator_scoring_function,
                temp_folder=optuna_folder,
                extra_config=extra_config,
                cuda=cuda,
            )
            tuner = optuna_result.tuner
            best_param = optuna_result.best_param
            self.best_params[model] = best_param
            self.optuna_results[model] = optuna_result

        num_jobs = max(num_jobs, num_parallel)
        repeat_temp_folder = os.path.join(temp_folder, "__repeat__")
        os.makedirs(repeat_temp_folder, exist_ok=True)
        model_configs = shallow_copy_dict(self.best_params)
        repeat_config = {
            "models": self.models,
            "model_configs": model_configs,
            "sequential": num_jobs <= 1,
            "predict_config": predict_config,
            "temp_folder": repeat_temp_folder,
            "num_repeat": num_final_repeat,
            "num_jobs": num_jobs,
            "cuda": cuda,
        }
        assert tuner is not None
        x, y, x_cv, y_cv = tuner.make_data()
        self.repeat_result = repeat_with(x, y, x_cv, y_cv, **repeat_config)
        self.pipelines = self.repeat_result.pipelines
        self.patterns = self.repeat_result.patterns
        self.data = self.repeat_result.data
        assert self.pipelines is not None
        assert self.patterns is not None

        keys = list(self.patterns.keys())
        scores = [self.studies[k].best_value for k in keys]
        keys_arr, weights_arr = map(np.array, [keys, scores])
        weights_arr /= weights_arr.sum()

        all_patterns = []
        pattern_weights = []
        self.weights_mapping = {}
        for key, weight in zip(keys, weights_arr.tolist()):
            patterns = self.patterns[key]
            num_patterns = len(patterns)
            avg_weight = weight / num_patterns
            all_patterns.extend(patterns)
            self.weights_mapping[key] = avg_weight
            pattern_weights.extend([avg_weight] * num_patterns)

        pattern_weights = np.array(pattern_weights, np.float32)
        self.pattern = Ensemble.stacking(all_patterns, pattern_weights=pattern_weights)

        return self

    def pack(
        self,
        export_folder: str,
        *,
        verbose: bool = False,
        use_tqdm: bool = True,
        compress: bool = True,
        retain_data: bool = False,
        remove_original: bool = True,
    ) -> "Auto":
        if self.pipelines is None:
            raise ValueError("`pipelines` are not yet generated")
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            Saving.prepare_folder(self, export_folder)
            # data
            if self.data is None:
                raise ValueError("`data` is not generated yet")
            with open(os.path.join(export_folder, self.data_protocol_file), "w") as f:
                f.write(self.data.__identifier__)  # type: ignore
            data_folder = os.path.join(export_folder, self.data_folder)
            self.data.save(
                data_folder,
                retain_data=retain_data,
                compress=False,
            )
            # weights
            weights_file = os.path.join(export_folder, self.weights_mapping_file)
            with open(weights_file, "w") as f:
                json.dump(self.weights_mapping, f)
            # core
            iterator = self.models
            if use_tqdm:
                iterator = tqdm(iterator, "pack")
            for model in iterator:
                pipelines = self.pipelines[model]
                model_folder = os.path.join(export_folder, model)
                Pack.pack_multiple(pipelines, model_folder, verbose=verbose)
            if compress:
                Saving.compress(abs_folder, remove_original=remove_original)
        return self

    @classmethod
    def unpack(
        cls,
        export_folder: str,
        device: Union[str, torch.device] = "cpu",
        *,
        compress: bool = True,
        use_tqdm: bool = True,
        use_tqdm_in_predictor: bool = False,
        **predict_kwargs: Any,
    ) -> UnPacked:
        patterns = []
        pattern_weights = []
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(
                export_folder,
                compress,
                remove_extracted=True,
            ):
                # data
                protocol_path = os.path.join(export_folder, cls.data_protocol_file)
                with open(protocol_path, "r") as f:
                    data_protocol = f.read().strip()
                data_folder = os.path.join(export_folder, cls.data_folder)
                data = DataProtocol.get(data_protocol).load(data_folder, compress=False)
                predictors = {}
                # weights
                weights_file = os.path.join(export_folder, cls.weights_mapping_file)
                with open(weights_file, "r") as f:
                    weights_mapping = json.load(f)
                # core
                iterator = [
                    stuff
                    for stuff in os.listdir(export_folder)
                    if stuff != cls.data_folder
                    and stuff != cls.weights_mapping_file
                    and stuff != cls.data_protocol_file
                ]
                if use_tqdm:
                    iterator = tqdm(iterator, "unpack")
                for model in iterator:
                    local_predictors = []
                    model_folder = os.path.join(export_folder, model)
                    for sub_folder in os.listdir(model_folder):
                        sub_folder = os.path.join(model_folder, sub_folder)
                        local_predictor = Pack.get_predictor(
                            sub_folder,
                            device,
                            data=data,
                            compress=False,
                            use_tqdm=use_tqdm_in_predictor,
                        )
                        local_predictors.append(local_predictor)
                        patterns.append(local_predictor.to_pattern(**predict_kwargs))
                        pattern_weights.append(weights_mapping[model])
                    predictors[model] = local_predictors

        pattern_weights = np.array(pattern_weights, np.float32)
        pattern = Ensemble.stacking(patterns, pattern_weights=pattern_weights)

        return UnPacked(pattern, predictors)

    # visualization

    def plot_param_importances(
        self,
        model: str,
        evaluator: Optional[BaseImportanceEvaluator] = None,
        params: Optional[List[str]] = None,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_param_importances(self.studies[model], evaluator, params)
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "param_importances.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_contour(
        self,
        model: str,
        params: Optional[List[str]] = None,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_contour(self.studies[model], params)
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "contour.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_parallel_coordinate(
        self,
        model: str,
        params: Optional[List[str]] = None,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_parallel_coordinate(self.studies[model], params)
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "parallel_coordinate.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_slice(
        self,
        model: str,
        params: Optional[List[str]] = None,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_slice(self.studies[model], params)
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "slice.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_optimization_history(
        self,
        model: str,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_optimization_history(self.studies[model])
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "optimization_history.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_intermediate_values(
        self,
        model: str,
        export_folder: Optional[str] = None,
    ) -> Figure:
        fig = vis.plot_intermediate_values(self.studies[model])
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "intermediate_values.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig

    def plot_edf(self, model: str, export_folder: str = None) -> Figure:
        fig = vis.plot_edf(self.studies[model])
        if export_folder is not None:
            os.makedirs(export_folder, exist_ok=True)
            html_path = os.path.join(export_folder, "edf.html")
            with open(html_path, "w") as f:
                f.write(fig.to_html())
        return fig


__all__ = ["Auto"]



from typing import Any
from typing import Dict

from ....configs import Configs


@Configs.register("identity", "default")
class DefaultIdentityConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return {}


@Configs.register("identity_ts", "default")
class DefaultIdentityTSConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return {}


__all__ = ["DefaultIdentityConfig", "DefaultIdentityTSConfig"]



import cflearn

from ._utils import get_info


if __name__ == "__main__":
    info = get_info()
    kwargs = info.kwargs
    increment_kwargs = info.increment_kwargs
    data_list = info.data_list
    assert data_list is not None
    sample_weights = info.meta.get("sample_weights")
    model = kwargs.pop("model")
    m = cflearn.make(model, kwargs, increment_kwargs)
    m.fit(*data_list, sample_weights=sample_weights)
    compress = info.meta.get("compress", True)
    cflearn.save(m, saving_folder=info.workplace, compress=compress)



import os
import json
import torch

import numpy as np

from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from functools import partial
from cftool.ml import ModelPattern
from cftool.misc import register_core
from cftool.misc import shallow_copy_dict
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import LoggingMixin
from mlflow.pyfunc import PythonModel
from mlflow.pyfunc import PythonModelContext

from ..types import data_type
from ..types import np_dict_type
from ..pipeline import Pipeline
from ..protocol import DataProtocol
from ..inference import ONNX
from ..inference import Inference
from ..inference import PreProcessor


class Predictor:
    def __init__(
        self,
        onnx_config: Dict[str, Any],
        preprocessor_folder: str,
        device: Union[str, torch.device] = "cpu",
        *,
        data: Optional[DataProtocol] = None,
        compress: bool = True,
        use_tqdm: bool = False,
    ):
        self.device = device
        preprocessor = PreProcessor.load(
            preprocessor_folder,
            data=data,
            compress=compress,
        )
        self.inference = Inference(
            preprocessor,
            onnx_config=onnx_config,
            use_tqdm=use_tqdm,
        )

    def __str__(self) -> str:
        return f"Predictor({self.inference})"

    __repr__ = __str__

    def predict(
        self,
        x: data_type,
        batch_size: int = 256,
        *,
        contains_labels: bool = False,
        **kwargs: Any,
    ) -> np_dict_type:
        loader = self.inference.preprocessor.make_inference_loader(
            x,
            self.device,
            batch_size,
            is_onnx=self.inference.onnx is not None,
            contains_labels=contains_labels,
        )
        kwargs = shallow_copy_dict(kwargs)
        kwargs["contains_labels"] = contains_labels
        return self.inference.predict(loader, **shallow_copy_dict(kwargs))

    def predict_prob(
        self,
        x: data_type,
        batch_size: int = 256,
        *,
        contains_labels: bool = False,
        **kwargs: Any,
    ) -> np_dict_type:
        kwargs = shallow_copy_dict(kwargs)
        kwargs["returns_probabilities"] = True
        return self.predict(
            x,
            batch_size,
            contains_labels=contains_labels,
            **shallow_copy_dict(kwargs),
        )

    def to_pattern(self, **kwargs: Any) -> ModelPattern:
        kwargs = shallow_copy_dict(kwargs)
        predict = partial(self.predict, **shallow_copy_dict(kwargs))
        kwargs["returns_probabilities"] = True
        predict_prob = partial(self.predict, **shallow_copy_dict(kwargs))
        return ModelPattern(predict_method=predict, predict_prob_method=predict_prob)


class Pack(LoggingMixin):
    def __init__(self, export_folder: str, *, loading: bool):
        if not loading:
            Saving.prepare_folder(self, export_folder)
        self.export_folder = os.path.abspath(export_folder)

    @property
    def onnx_path(self) -> str:
        return os.path.join(self.export_folder, "m.onnx")

    @property
    def onnx_output_names_path(self) -> str:
        return os.path.join(self.export_folder, "output_names.json")

    @property
    def output_probabilities_path(self) -> str:
        return os.path.join(self.export_folder, "output_probabilities.txt")

    @property
    def binary_config_path(self) -> str:
        return os.path.join(self.export_folder, "binary_config.json")

    @property
    def preprocessor_folder(self) -> str:
        return os.path.join(self.export_folder, "preprocessor")

    @classmethod
    def pack(
        cls,
        pipeline: Pipeline,
        export_folder: str,
        *,
        verbose: bool = True,
        compress: bool = True,
        pack_data: bool = True,
        retain_data: bool = False,
        remove_original: bool = True,
        **kwargs: Any,
    ) -> None:
        kwargs = shallow_copy_dict(kwargs)
        kwargs["verbose"] = verbose
        instance = cls(export_folder, loading=False)
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            model = pipeline.model
            if model is None:
                raise ValueError("`model` is not generated yet")
            with model.export_context():
                onnx = ONNX(model=model)
                onnx.to_onnx(instance.onnx_path, **shallow_copy_dict(kwargs))
            with open(instance.onnx_output_names_path, "w") as f:
                json.dump(onnx.output_names, f)
            with open(instance.output_probabilities_path, "w") as f:
                f.write(str(int(model.output_probabilities)))
            pipeline.preprocessor.save(
                instance.preprocessor_folder,
                save_data=pack_data,
                retain_data=retain_data,
                compress=False,
            )
            with open(instance.binary_config_path, "w") as f:
                trainer = pipeline.trainer
                inference = trainer.inference
                if inference.binary_threshold is None:
                    trainer._generate_binary_threshold()
                json.dump(inference.binary_config, f)
            if compress:
                Saving.compress(abs_folder, remove_original=remove_original)

    @classmethod
    def pack_multiple(
        cls,
        pipelines: List[Pipeline],
        export_folder: str,
        *,
        verbose: bool = True,
    ) -> None:
        """ This method will simply pack onnx models and won't retain data """
        for i, pipeline in enumerate(pipelines):
            local_export_folder = os.path.join(export_folder, f"m_{i:04d}")
            Pack.pack(
                pipeline,
                local_export_folder,
                verbose=verbose,
                pack_data=False,
                compress=False,
            )

    @classmethod
    def get_predictor(
        cls,
        export_folder: str,
        device: Union[str, torch.device] = "cpu",
        *,
        data: Optional[DataProtocol] = None,
        compress: bool = True,
        use_tqdm: bool = False,
    ) -> Predictor:
        instance = cls(export_folder, loading=True)
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(
                export_folder,
                compress,
                remove_extracted=True,
                logging_mixin=instance,
            ):
                with open(instance.onnx_output_names_path, "r") as f:
                    output_names = json.load(f)
                with open(instance.output_probabilities_path, "r") as f:
                    output_probabilities = bool(int(f.read().strip()))
                onnx_config = {
                    "onnx_path": instance.onnx_path,
                    "output_names": output_names,
                    "output_probabilities": output_probabilities,
                }
                predictor = Predictor(
                    onnx_config,
                    instance.preprocessor_folder,
                    device,
                    data=data,
                    compress=False,
                    use_tqdm=use_tqdm,
                )
                with open(instance.binary_config_path, "r") as f:
                    cfg = json.load(f)
                    predictor.inference.inject_binary_config(cfg)
        return predictor


python_models: Dict[str, Type["PythonModelBase"]] = {}


class PythonModelBase(PythonModel, metaclass=ABCMeta):
    """ Should implement self.predictor """

    def predict(
        self,
        context: PythonModelContext,
        model_input: Any,
    ) -> np.ndarray:
        """
        model_inputs should be a pd.DataFrame which looks like:

        |    file    |   other keys   |
        |  "xxx.csv" |  other values  |

        """
        columns = model_input.columns
        data = model_input.values.ravel()
        kwargs = dict(zip(columns[1:], data[1:]))
        return self.predictor.predict(data[0], **kwargs)

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global python_models
        return register_core(name, python_models)


@PythonModelBase.register("pack")
class PackModel(PythonModelBase):
    def load_context(self, context: PythonModelContext) -> None:
        pack_folder = context.artifacts["pack_folder"]
        self.predictor = Pack.get_predictor(pack_folder, compress=False)


@PythonModelBase.register("pipeline")
class PipelineModel(PythonModelBase):
    def load_context(self, context: PythonModelContext) -> None:
        export_folder = context.artifacts["export_folder"]
        self.predictor = Pipeline.load(export_folder, compress=False)


__all__ = ["Pack", "PackModel", "PipelineModel"]



from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Callable
from cftool.misc import register_core

from ...types import tensor_dict_type


aggregator_dict: Dict[str, Type["AggregatorBase"]] = {}


class AggregatorBase(metaclass=ABCMeta):
    def __init__(self, **kwargs: Any):
        self.config = kwargs

    @abstractmethod
    def reduce(self, outputs: tensor_dict_type, **kwargs: Any) -> tensor_dict_type:
        """ requires returning the `predictions` key """

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global aggregator_dict
        return register_core(name, aggregator_dict)

    @classmethod
    def make(cls, name: str, **kwargs: Any) -> "AggregatorBase":
        return aggregator_dict[name](**kwargs)


__all__ = ["AggregatorBase"]



import os
import json
import argparse

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import NamedTuple
from cftool.misc import Saving
from cfdata.tabular import data_type

meta_config_name = "__meta__"
data_config_file = "__data__.json"


class Info(NamedTuple):
    workplace: str
    meta: Dict[str, Any]
    kwargs: Dict[str, Any]
    increment_kwargs: Dict[str, Any]
    data_list: Optional[List[data_type]]


def get_info(*, requires_data: bool = True) -> Info:
    parser = argparse.ArgumentParser()
    parser.add_argument("--config_folder", type=str)
    args = parser.parse_args()
    # common
    meta_config = Saving.load_dict(meta_config_name, args.config_folder)
    cuda = meta_config["cuda"]
    kwargs = meta_config["config"]
    workplace = meta_config["workplace"]
    increment_kwargs = meta_config["increment_config"]
    increment_kwargs["cuda"] = cuda
    increment_kwargs["logging_folder"] = os.path.join(workplace, "_logs")
    log_artifacts_key = "log_pipeline_to_artifacts"
    increment_kwargs[log_artifacts_key] = kwargs.get(log_artifacts_key, True)
    # data
    if not requires_data:
        data_list = None
    else:
        data_folder = increment_kwargs.get("data_folder")
        if data_folder is None:
            raise ValueError("`data_folder` should be provided")
        data_config_path = os.path.join(data_folder, data_config_file)
        keys = ["x", "y", "x_cv", "y_cv"]
        if os.path.isfile(data_config_path):
            with open(data_config_path, "r") as f:
                data_config = json.load(f)
            data_list = list(map(data_config.get, keys))
        else:
            data_list = []
            for key in keys:
                data_file = os.path.join(data_folder, f"{key}.npy")
                if not os.path.isfile(data_file):
                    data_list.append(None)
                else:
                    data_list.append(np.load(data_file))
    return Info(workplace, meta_config, kwargs, increment_kwargs, data_list)


__all__ = [
    "get_info",
]



import torch

from typing import Any
from typing import Dict

from ..base import ExtractorBase
from ...transform.core import Dimensions


rnn_dict = {"LSTM": torch.nn.LSTM, "GRU": torch.nn.GRU, "RNN": torch.nn.RNN}


@ExtractorBase.register("rnn")
class RNN(ExtractorBase):
    def __init__(
        self,
        in_flat_dim: int,
        dimensions: Dimensions,
        cell: str,
        cell_config: Dict[str, Any],
        num_layers: int = 1,
    ):
        super().__init__(in_flat_dim, dimensions)
        # rnn
        rnn_base = rnn_dict[cell]
        input_dimensions = [self.in_dim]
        self.hidden_size = cell_config["hidden_size"]
        input_dimensions += [self.hidden_size] * (num_layers - 1)
        self.rnn_list = torch.nn.ModuleList(
            [rnn_base(dim, **cell_config) for dim in input_dimensions]
        )

    @property
    def flatten_ts(self) -> bool:
        return False

    @property
    def out_dim(self) -> int:
        return self.hidden_size

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        for rnn in self.rnn_list:
            net, final_state = rnn(net, None)
        return net[..., -1, :]


__all__ = ["RNN"]



import os

import numpy as np
import matplotlib.pyplot as plt

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional
from cftool.misc import show_or_save

from ...types import data_type
from ...pipeline import Pipeline


class DDRPredictor:
    def __init__(self, ddr: Pipeline):
        self.m = ddr

    def mr(self, x: data_type) -> Dict[str, np.ndarray]:
        return self.m.predict(x, predict_median_residual=True, return_all=True)

    def cdf(
        self,
        x: data_type,
        y: data_type,
        *,
        get_pdf: bool = False,
    ) -> Dict[str, np.ndarray]:
        predictions = self.m.predict(
            x,
            y=y,
            use_grad=get_pdf,
            requires_recover=False,
            predict_pdf=get_pdf,
            predict_cdf=True,
            return_all=True,
        )
        return predictions

    def quantile(
        self,
        x: data_type,
        q: Union[float, List[float]],
        *,
        recover: bool = True,
    ) -> Dict[str, np.ndarray]:
        return self.m.predict(
            x,
            q=q,
            requires_recover=recover,
            predict_quantiles=True,
            return_all=True,
        )


class DDRVisualizer:
    def __init__(
        self,
        ddr: Pipeline,
        dpi: int = 200,
        figsize: Tuple[int, int] = (8, 6),
    ):
        self.m = ddr
        self.dpi = dpi
        self.figsize = figsize
        self.predictor = DDRPredictor(ddr)

    def _prepare_base_figure(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_base: np.ndarray,
        mean: Optional[np.ndarray],
        median: Optional[np.ndarray],
        indices: np.ndarray,
        title: str,
    ) -> plt.Figure:
        figure = plt.figure(figsize=self.figsize, dpi=self.dpi)
        plt.title(title)
        plt.scatter(x[indices], y[indices], color="gray", s=15)
        if mean is not None:
            plt.plot(x_base.ravel(), mean.ravel(), label="mean")
        if median is not None:
            plt.plot(x_base.ravel(), median.ravel(), label="median")
        return figure

    @staticmethod
    def _render_figure(
        x_min: np.ndarray,
        x_max: np.ndarray,
        y_min: float,
        y_max: float,
        y_padding: float,
    ) -> None:
        plt.xlim(x_min, x_max)
        plt.ylim(y_min - 0.5 * y_padding, y_max + 0.5 * y_padding)
        plt.legend()

    def visualize(
        self,
        x: np.ndarray,
        y: np.ndarray,
        export_path: Optional[str],
        *,
        median_residual: bool = False,
        mul_affine: bool = False,
        cdf_logit_mul: bool = False,
        q_batch: Optional[np.ndarray] = None,
        y_batch: Optional[np.ndarray] = None,
        to_pdf: bool = False,
        **kwargs: Any,
    ) -> None:
        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)
        padding, dense = kwargs.get("padding", 1.0), kwargs.get("dense", 400)
        indices = np.random.permutation(len(x))[:dense]
        x_padding = max(abs(x_min), abs(x_max)) * padding
        y_padding = max(abs(y_min), abs(y_max)) * padding
        x_min -= x_padding
        x_max += x_padding
        x_base = np.linspace(x_min, x_max, dense)[..., None]
        model = self.m.model
        assert model is not None
        mean = None
        median = self.m.predict(x_base)
        fig = self._prepare_base_figure(x, y, x_base, mean, median, indices, "")
        render_args = x_min, x_max, y_min, y_max, y_padding
        # median residual
        if median_residual:
            residuals = self.predictor.mr(x_base)
            pos, neg = map(residuals.get, ["mr_pos", "mr_neg"])
            plt.plot(x_base.ravel(), pos, label="pos_median_residual")
            plt.plot(x_base.ravel(), neg, label="neg_median_residual")
            DDRVisualizer._render_figure(*render_args)
        # quantile curves
        if q_batch is not None and model.fetch_q:
            q_list = q_batch.tolist()
            if not mul_affine:
                quantile_curves = self.predictor.quantile(x_base, q_list)["quantiles"]
                for q, quantile_curve in zip(q_list, quantile_curves.T):
                    plt.plot(x_base.ravel(), quantile_curve, label=f"quantile {q:4.2f}")
                DDRVisualizer._render_figure(*render_args)
            else:
                med_predictions = self.predictor.quantile(x_base, q_list, recover=False)
                med_mul_cat = med_predictions["med_mul"]
                for q, med_mul in zip(q_batch, med_mul_cat.T):
                    plt.plot(x_base.ravel(), med_mul, label=f"med_mul {q:4.2f}")
                DDRVisualizer._render_figure(*render_args)
        # cdf curves
        if y_batch is not None and model.fetch_cdf:
            y_abs_max = np.abs(y).max()
            ratios = y_batch
            anchors = [ratio * (y_max - y_min) + y_min for ratio in y_batch]
            if not cdf_logit_mul:
                for ratio, anchor in zip(ratios, anchors):
                    anchor_line = np.full(len(x_base), anchor)
                    predictions = self.predictor.cdf(x_base, anchor, get_pdf=to_pdf)
                    if not to_pdf:
                        cdf = predictions["cdf"] * y_abs_max
                        plt.plot(x_base.ravel(), cdf, label=f"cdf {ratio:4.2f}")
                    else:
                        pdf = predictions["pdf"]
                        pdf = pdf * (y_abs_max / max(np.abs(pdf).max(), 1e-8))
                        plt.plot(x_base.ravel(), pdf, label=f"pdf {ratio:4.2f}")
                    plt.plot(
                        x_base.ravel(),
                        anchor_line,
                        label=f"anchor {ratio:4.2f}",
                        color="gray",
                    )
                DDRVisualizer._render_figure(*render_args)
            else:
                for ratio, anchor in zip(ratios, anchors):
                    cdf_predictions = self.predictor.cdf(x_base, anchor)
                    cdf_logit_mul = cdf_predictions["cdf_logit_mul"]
                    label = f"cdf_logit_mul {ratio:4.2f}"
                    plt.plot(x_base.ravel(), cdf_logit_mul, label=label)
                DDRVisualizer._render_figure(*render_args)
        show_or_save(export_path, fig)

    def visualize_multiple(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_base: np.ndarray,
        y_matrix: np.ndarray,
        export_folder: str,
        *,
        q_batch: Optional[np.ndarray] = None,
        y_batch: Optional[np.ndarray] = None,
    ) -> None:
        y_min, y_max = y.min(), y.max()
        y_diff = y_max - y_min

        def _plot(
            prefix: str,
            num: int,
            y_true: Optional[np.ndarray],
            predictions: np.ndarray,
            anchor_line_: Optional[np.ndarray],
        ) -> None:
            plt.figure(figsize=self.figsize, dpi=self.dpi)
            plt.title(f"{prefix} {num:6.4f}")
            plt.scatter(x[:200], y[:200], color="gray", s=15)
            if y_true is not None:
                plt.plot(x_base, y_true, label="target")
            plt.plot(x_base, predictions, label=f"ddr_prediction")
            if anchor_line_ is not None:
                plt.plot(x_base.ravel(), anchor_line_, color="gray")
            plt.legend()
            show_or_save(os.path.join(export_folder, f"{prefix}_{num:4.2f}.png"))

        if q_batch is not None:
            q_list = q_batch.tolist()
            yq_predictions = self.predictor.quantile(x_base, q_list)["quantiles"]
            for q, yq_pred in zip(q_list, yq_predictions.T):
                yq = np.percentile(y_matrix, int(100 * q), axis=1)
                _plot("quantile", q, yq, yq_pred, None)
        if y_batch is not None:
            anchors = [ratio * (y_max - y_min) + y_min for ratio in y_batch]
            for anchor in anchors:
                anchor_line = np.full(len(x_base), anchor)
                yd = np.mean(y_matrix <= anchor, axis=1) * y_diff + y_min
                yd_pred = self.predictor.cdf(x_base, anchor)["cdf"]
                yd_pred = yd_pred * y_diff + y_min
                _plot("cdf", anchor, yd, yd_pred, anchor_line)


__all__ = ["DDRPredictor", "DDRVisualizer"]



from .ddr import *
from .ndt import *
from .nnb import *
from .rnn import *
from .wnd import *
from .fcnn import *
from .linear import *
from .tree_dnn import *
from .transformer import *
from .base import ModelBase



import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Callable
from cftool.misc import register_core

from .types import losses_type
from .types import tensor_dict_type
from .misc.toolkit import to_torch


loss_dict: Dict[str, Type["LossBase"]] = {}


class LossBase(nn.Module, metaclass=ABCMeta):
    def __init__(self, config: Dict[str, Any], reduction: str = "mean"):
        super().__init__()
        self.config = config
        self._init_config(config)
        self._reduction = reduction

    def _init_config(self, config: Dict[str, Any]) -> None:
        pass

    def _reduce(self, losses: torch.Tensor) -> torch.Tensor:
        if self._reduction == "none":
            return losses
        if self._reduction == "mean":
            return losses.mean()
        if self._reduction == "sum":
            return losses.sum()
        raise NotImplementedError(f"reduction '{self._reduction}' is not implemented")

    @abstractmethod
    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        # return losses without reduction
        pass

    def forward(
        self,
        predictions: tensor_dict_type,
        target: torch.Tensor,
    ) -> losses_type:
        losses = self._core(predictions, target)
        if isinstance(losses, torch.Tensor):
            return self._reduce(losses)
        return {k: self._reduce(v) for k, v in losses.items()}

    @classmethod
    def make(
        cls,
        name: str,
        config: Dict[str, Any],
        reduction: str = "mean",
    ) -> "LossBase":
        return loss_dict[name](config, reduction)

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global loss_dict
        return register_core(name, loss_dict)


@LossBase.register("mae")
class MAELoss(LossBase):
    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        return F.l1_loss(forward_results["predictions"], target, reduction="none")


@LossBase.register("mse")
class MSELoss(LossBase):
    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        return F.mse_loss(forward_results["predictions"], target, reduction="none")


@LossBase.register("quantile")
class Quantile(LossBase):
    def _init_config(self, config: Dict[str, Any]) -> None:
        q = config.get("q")
        if q is None:
            raise ValueError("'q' should be provided in Quantile loss")
        if isinstance(q, float):
            self.register_buffer("q", torch.tensor([q], torch.float32))
        else:
            q = np.asarray(q, np.float32).reshape([1, -1])
            self.register_buffer("q", to_torch(q))

    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        quantile_error = target - forward_results["quantiles"]
        neg_errors = self.q * quantile_error  # type: ignore
        pos_errors = (self.q - 1) * quantile_error  # type: ignore
        quantile_losses = torch.max(neg_errors, pos_errors)
        return quantile_losses.mean(1, keepdim=True)


@LossBase.register("cross_entropy")
class CrossEntropy(LossBase):
    @staticmethod
    def _get_stat(
        predictions: torch.Tensor,
        target: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        logits_mat = predictions.view(-1, predictions.shape[-1])
        log_prob_mat = F.log_softmax(logits_mat, dim=1)
        nll_losses = -log_prob_mat.gather(dim=1, index=target)
        return log_prob_mat, nll_losses

    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        return self._get_stat(forward_results["predictions"], target)[1]


@LossBase.register("label_smooth_cross_entropy")
class LabelSmoothCrossEntropy(LossBase):
    def _init_config(self, config: Dict[str, Any]) -> None:
        self._eps = config.setdefault("eps", 0.1)

    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        predictions = forward_results["predictions"]
        log_prob_mat, nll_losses = CrossEntropy._get_stat(predictions, target)
        smooth_losses = -log_prob_mat.sum(dim=1, keepdim=True)
        eps = self._eps / log_prob_mat.shape[-1]
        return (1.0 - self._eps) * nll_losses + eps * smooth_losses


@LossBase.register("focal")
class FocalLoss(LossBase):
    def _init_config(self, config: Dict[str, Any]) -> None:
        self._input_logits = config.setdefault("input_logits", True)
        self._eps = config.setdefault("eps", 1e-6)
        self._gamma = config.setdefault("gamma", 2.0)
        alpha = config.setdefault("alpha", None)
        if isinstance(alpha, (int, float)):
            alpha = [alpha, 1 - alpha]
        elif isinstance(alpha, (list, tuple)):
            alpha = list(alpha)
        if alpha is None:
            self.alpha = None
        else:
            self.register_buffer("alpha", to_torch(np.array(alpha, np.float32)))

    def _core(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        **kwargs: Any,
    ) -> losses_type:
        predictions = forward_results["predictions"]
        if not self._input_logits:
            prob_mat = predictions.view(-1, predictions.shape[-1]) + self._eps
        else:
            logits_mat = predictions.view(-1, predictions.shape[-1])
            prob_mat = F.softmax(logits_mat, dim=1) + self._eps
        gathered_prob_flat = prob_mat.gather(dim=1, index=target).view(-1)
        gathered_log_prob_flat = gathered_prob_flat.log()
        if self.alpha is not None:
            alpha_target = self.alpha.gather(dim=0, index=target.view(-1))
            gathered_log_prob_flat = gathered_log_prob_flat * alpha_target
        return -gathered_log_prob_flat * (1 - gathered_prob_flat) ** self._gamma


__all__ = ["LossBase"]



from .base import head_dict
from .base import HeadBase
from .base import HeadConfigs
from .ddr import DDRHead
from .dndf import DNDFHead
from .fcnn import FCNNHead
from .linear import LinearHead
from .tree_stack import TreeStackHead
from .traditional import NNBMNBHead
from .traditional import NNBNormalHead


__all__ = [
    "head_dict",
    "HeadBase",
    "HeadConfigs",
]



import os

import numpy as np

from typing import *
from tqdm.autonotebook import tqdm
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import LoggingMixin
from cftool.ml.utils import patterns_type
from cftool.ml.utils import Comparer
from cftool.ml.utils import Estimator
from cftool.ml.utils import ModelPattern
from cftool.ml.utils import EnsemblePattern
from cfdata.tabular import task_type_type
from cfdata.tabular import parse_task_type
from cfdata.tabular import TaskTypes

from ..dist import inject_distributed_tqdm_kwargs
from ..dist import Task
from ..dist import Experiment
from ..dist import ExperimentResults
from ..types import data_type
from ..types import general_config_type
from ..configs import _parse_config
from ..trainer import Trainer
from ..trainer import TrainerCallback
from ..trainer import IntermediateResults
from ..pipeline import Pipeline
from ..pipeline import ExternalPipelineProtocol
from ..protocol import DataProtocol
from ..misc._api import _to_saving_path
from ..misc._api import _make_saving_path
from ..misc._api import _fetch_saving_paths
from ..misc.toolkit import to_2d
from ..misc.toolkit import inject_mlflow_stuffs
from ..models.base import model_dict
from ..models.base import PipeConfig


def make(
    model: str = "fcnn",
    config: general_config_type = None,
    increment_config: general_config_type = None,
    **kwargs: Any,
) -> Pipeline:
    kwargs["model"] = model
    parsed_config = update_dict(_parse_config(config), kwargs)
    parsed_increment_config = _parse_config(increment_config)
    return Pipeline.make(parsed_config, parsed_increment_config)


def make_external(
    protocol: ExternalPipelineProtocol,
    config: general_config_type = None,
    increment_config: general_config_type = None,
    **kwargs: Any,
) -> Pipeline:
    m = make("", config, increment_config, **kwargs)
    m.custom_fit = protocol.fit
    m.custom_predict = protocol.predict  # type: ignore
    return m


class _PipeConfigManager:
    def __init__(
        self,
        pipes: Dict[str, PipeConfig],
        meta_configs: Dict[str, Dict[str, Any]],
        pipe: str,
    ):
        self.pipe_config = pipes[pipe]
        self.meta_config = meta_configs.setdefault(pipe, {})
        self.pipes, self.pipe = pipes, pipe

    @property
    def transform_config(self) -> Dict[str, Any]:
        return self.meta_config.setdefault("transform", {})

    @transform_config.setter
    def transform_config(self, value: Dict[str, Any]) -> None:
        self.meta_config["transform"] = value

    @property
    def extractor_config(self) -> Dict[str, Any]:
        return self.meta_config.setdefault("extractor", {})

    @extractor_config.setter
    def extractor_config(self, value: Dict[str, Any]) -> None:
        self.meta_config["extractor"] = value

    @property
    def head_config(self) -> Dict[str, Any]:
        return self.meta_config.setdefault("head", {})

    @head_config.setter
    def head_config(self, value: Dict[str, Any]) -> None:
        self.meta_config["head"] = value

    def replace(self, **kwargs: Any) -> None:
        pipe_config_dict = self.pipe_config._asdict()
        pipe_config_dict.update(kwargs)
        self.pipe_config = PipeConfig(**pipe_config_dict)
        self.pipes[self.pipe] = self.pipe_config


class ModelConfig:
    def __init__(self, name: str):
        self.name = name
        base = model_dict[name]
        self.registered_pipes = base.registered_pipes
        self.registered_meta_configs = base.registered_meta_configs or {}
        base.registered_meta_configs = self.registered_meta_configs

    def switch(self, pipe: Optional[str] = None) -> _PipeConfigManager:
        if self.registered_pipes is None:
            raise ValueError(f"no pipes are registered for {self.name}")
        if pipe is None:
            if len(self.registered_pipes) > 1:
                pipe_str = ", ".join(sorted(self.registered_pipes))
                raise ValueError(
                    "`pipe` must be specified when more than one pipe is registered "
                    f"(current registered pipes: {pipe_str})"
                )
            pipe = list(self.registered_pipes.keys())[0]
        return _PipeConfigManager(
            self.registered_pipes,
            self.registered_meta_configs,
            pipe,
        )


def make_from(
    identifier: str = "cflearn",
    saving_folder: Optional[str] = None,
    kwargs_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    cuda: Optional[Union[int, str]] = None,
    *,
    new_model: Optional[str] = None,
    load_increment: bool = True,
    increment_config: general_config_type = None,
    model_mapping: Optional[Dict[str, str]] = None,
) -> Dict[str, List[Pipeline]]:
    saving_folder = os.path.abspath(saving_folder or "./")
    if new_model is None:
        model_mapping = None
    else:
        paths_dict = _fetch_saving_paths(identifier, saving_folder)
        all_paths: List[str] = sum(paths_dict.values(), [])
        if len(all_paths) > 1:
            raise ValueError("more than 1 model is detected")
        pretrain_model = list(paths_dict.keys())[0]
        model_mapping = {pretrain_model: new_model}

    def _core() -> Pipeline:
        compress = os.path.isfile(f"{path}.zip")
        with lock_manager(saving_folder, [path]):
            with Saving.compress_loader(path, compress):
                config_bundle = Saving.load_dict(Pipeline.config_bundle_name, path)
        kwargs = config_bundle["config"]
        if increment_config is None:
            increment_kwargs = {}
        else:
            increment_kwargs = _parse_config(increment_config)
        if load_increment:
            loaded_inc_config = config_bundle["increment_config"]
            increment_kwargs = update_dict(increment_kwargs, loaded_inc_config)
        increment_kwargs.setdefault("binary_config", config_bundle["binary_config"])
        if model_mapping is not None:
            kwargs["model"] = model_mapping[model]
        if cuda is not None:
            increment_kwargs["cuda"] = cuda
        kwargs.pop("logging_folder", None)
        if kwargs.get("mlflow_config") is not None:
            kwargs["mlflow_config"] = {}
        if kwargs_callback is not None:
            kwargs_callback(kwargs)
        return make(config=kwargs, increment_config=increment_kwargs)

    ms: Dict[str, List[Pipeline]] = {}
    paths_dict = _fetch_saving_paths(identifier, saving_folder)
    for model, paths in paths_dict.items():
        for path in paths:
            ms.setdefault(model, []).append(_core())
    return ms


def finetune(
    x: data_type,
    y: data_type = None,
    x_cv: data_type = None,
    y_cv: data_type = None,
    *,
    new_model: Optional[str] = None,
    strict: bool = True,
    load_increment: bool = True,
    identifier: str = "cflearn",
    pretrain_folder: Optional[str] = None,
    increment_config: general_config_type = None,
    kwargs_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    sample_weights: Optional[np.ndarray] = None,
    cuda: Optional[Union[int, str]] = None,
) -> Pipeline:
    ms = make_from(
        identifier,
        pretrain_folder,
        kwargs_callback,
        cuda,
        new_model=new_model,
        load_increment=load_increment,
        increment_config=increment_config,
    )
    m = list(ms.values())[0][0]
    m.fit(
        x,
        y,
        x_cv,
        y_cv,
        pretrain_strict=strict,
        pretrain_folder=pretrain_folder,
        pretrain_identifier=identifier,
        state_dict_callback=state_dict_callback,
        sample_weights=sample_weights,
    )
    return m


pipelines_type = Union[
    Pipeline,
    List[Pipeline],
    Dict[str, Pipeline],
    Dict[str, List[Pipeline]],
]


def _to_pipelines(pipelines: pipelines_type) -> Dict[str, List[Pipeline]]:
    if isinstance(pipelines, dict):
        pipeline_dict = {}
        for key, value in pipelines.items():
            if isinstance(value, list):
                pipeline_dict[key] = value
            else:
                pipeline_dict[key] = [value]
    else:
        if not isinstance(pipelines, list):
            pipelines = [pipelines]
        pipeline_dict = {}
        for pipeline in pipelines:
            assert pipeline.model is not None
            key = pipeline.model.__identifier__
            pipeline_dict.setdefault(key, []).append(pipeline)
    return pipeline_dict


def evaluate(
    x: data_type,
    y: data_type = None,
    *,
    contains_labels: bool = True,
    pipelines: Optional[pipelines_type] = None,
    predict_config: Optional[Dict[str, Any]] = None,
    metrics: Optional[Union[str, List[str]]] = None,
    metric_configs: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
    other_patterns: Optional[Dict[str, patterns_type]] = None,
    comparer_verbose_level: Optional[int] = 1,
) -> Comparer:
    if not contains_labels:
        msg = "`cflearn.evaluate` must be called with `contains_labels = True`"
        raise ValueError(msg)

    patterns = {}
    if pipelines is None:
        if y is None:
            raise ValueError("either `pipelines` or `y` should be provided")
        if metrics is None:
            raise ValueError("either `pipelines` or `metrics` should be provided")
        if metric_configs is None:
            metric_configs = [{} for _ in range(len(metrics))]
        if other_patterns is None:
            raise ValueError(
                "either `pipelines` or `other_patterns` should be provided"
            )
    else:
        pipelines = _to_pipelines(pipelines)
        if predict_config is None:
            predict_config = {}
        predict_config.setdefault("contains_labels", contains_labels)
        for name, pipeline_list in pipelines.items():
            first_pipeline = pipeline_list[0]
            data = first_pipeline.data
            if y is not None:
                y = to_2d(y)
            else:
                if not isinstance(x, str):
                    raise ValueError("`x` should be str when `y` is not provided")
                x, y = data.read_file(x, contains_labels=contains_labels)
                y = data.transform(x, y).y
            if metrics is None:
                metrics = [
                    k
                    for k, v in first_pipeline.trainer.metrics.items()
                    if v is not None
                ]
            if metric_configs is None:
                metric_configs = [
                    v.config
                    for k, v in first_pipeline.trainer.metrics.items()
                    if v is not None
                ]
            patterns[name] = [
                pipeline.to_pattern(**predict_config) for pipeline in pipeline_list
            ]
    if other_patterns is not None:
        for other_name in other_patterns.keys():
            if other_name in patterns:
                print(
                    f"{LoggingMixin.warning_prefix}'{other_name}' is found in "
                    "`other_patterns`, it will be overwritten"
                )
        update_dict(other_patterns, patterns)

    if isinstance(metrics, list):
        metrics_list = metrics
    else:
        assert isinstance(metrics, str)
        metrics_list = [metrics]
    if isinstance(metric_configs, list):
        metric_configs_list = metric_configs
    else:
        assert isinstance(metric_configs, dict)
        metric_configs_list = [metric_configs]

    estimators = [
        Estimator(metric, metric_config=metric_config)
        for metric, metric_config in zip(metrics_list, metric_configs_list)
    ]
    comparer = Comparer(patterns, estimators)
    comparer.compare(x, y, verbose_level=comparer_verbose_level)
    return comparer


def save(
    pipelines: pipelines_type,
    identifier: str = "cflearn",
    saving_folder: Optional[str] = None,
    *,
    compress: bool = True,
) -> Dict[str, List[Pipeline]]:
    pipeline_dict = _to_pipelines(pipelines)
    saving_path = _to_saving_path(identifier, saving_folder)
    for name, pipeline_list in pipeline_dict.items():
        for i, pipeline in enumerate(pipeline_list):
            pipeline.save(
                _make_saving_path(i, name, saving_path, True),
                compress=compress,
            )
    return pipeline_dict


def load(
    identifier: str = "cflearn",
    saving_folder: Optional[str] = None,
    *,
    compress: bool = True,
) -> Dict[str, List[Pipeline]]:
    paths = _fetch_saving_paths(identifier, saving_folder)
    pipelines = {
        k: [Pipeline.load(v, compress=compress) for v in v_list]
        for k, v_list in paths.items()
    }
    if not pipelines:
        raise ValueError(
            f"'{identifier}' models not found with `saving_folder`={saving_folder}"
        )
    return pipelines


def task_loader(saving_folder: str, compress: bool = True) -> Pipeline:
    return list(load(saving_folder=saving_folder, compress=compress).values())[0][0]


def load_experiment_results(results: ExperimentResults) -> Dict[str, List[Pipeline]]:
    pipelines_dict: Dict[str, Dict[int, Pipeline]] = {}
    for workplace, workplace_key in zip(results.workplaces, results.workplace_keys):
        pipeline = task_loader(workplace)
        model, str_i = workplace_key
        pipelines_dict.setdefault(model, {})[int(str_i)] = pipeline
    return {k: [v[i] for i in sorted(v)] for k, v in pipelines_dict.items()}


class RepeatResult(NamedTuple):
    data: Optional[DataProtocol]
    experiment: Optional[Experiment]
    pipelines: Optional[Dict[str, List[Pipeline]]]
    patterns: Optional[Dict[str, List[ModelPattern]]]

    @property
    def trainers(self) -> Optional[Dict[str, List[Trainer]]]:
        if self.pipelines is None:
            return None
        return {k: [m.trainer for m in v] for k, v in self.pipelines.items()}

    @property
    def final_results(self) -> Optional[Dict[str, List[IntermediateResults]]]:
        trainers = self.trainers
        if trainers is None:
            return None
        final_results_dict: Dict[str, List[IntermediateResults]] = {}
        for k, v in trainers.items():
            local_results = final_results_dict.setdefault(k, [])
            for trainer in v:
                final_results = trainer.final_results
                if final_results is None:
                    raise ValueError(f"training of `Trainer` ({trainer}) is corrupted")
                local_results.append(final_results)
        return final_results_dict


# If `x_cv` is not provided, then `KRandom` split will be performed
def repeat_with(
    x: data_type,
    y: data_type = None,
    x_cv: data_type = None,
    y_cv: data_type = None,
    *,
    models: Union[str, List[str]] = "fcnn",
    model_configs: Optional[Dict[str, Dict[str, Any]]] = None,
    predict_config: Optional[Dict[str, Any]] = None,
    sequential: Optional[bool] = None,
    num_jobs: int = 1,
    num_repeat: int = 5,
    temp_folder: str = "__tmp__",
    return_patterns: bool = True,
    compress: bool = True,
    use_tqdm: bool = True,
    **kwargs: Any,
) -> RepeatResult:
    if isinstance(models, str):
        models = [models]

    kwargs = shallow_copy_dict(kwargs)
    kwargs.setdefault("trigger_logging", False)
    kwargs["verbose_level"] = 0

    if sequential is None:
        sequential = num_jobs <= 1
    if model_configs is None:
        model_configs = {}

    def fetch_config(i_: int, model_: str) -> Dict[str, Any]:
        local_kwargs = shallow_copy_dict(kwargs)
        assert model_configs is not None
        local_model_config = model_configs.setdefault(model_, {})
        fetched = update_dict(shallow_copy_dict(local_model_config), local_kwargs)
        inject_mlflow_stuffs(model_, config=fetched, run_name=f"run_{i_}")
        return shallow_copy_dict(fetched)

    pipelines_dict: Optional[Dict[str, List[Pipeline]]] = None
    if sequential:
        experiment = None
        kwargs["tqdm_position"] = 2
        if not return_patterns:
            print(
                f"{LoggingMixin.warning_prefix}`return_patterns` should be "
                "True when `sequential` is True, because patterns "
                "will always be generated"
            )
            return_patterns = True
        pipelines_dict = {}
        if not use_tqdm:
            iterator = models
        else:
            iterator = tqdm(models, total=len(models), position=0)
        for model in iterator:
            local_pipelines = []
            sub_iterator = range(num_repeat)
            if use_tqdm:
                sub_iterator = tqdm(
                    sub_iterator,
                    total=num_repeat,
                    position=1,
                    leave=False,
                )
            for i in sub_iterator:
                local_config = fetch_config(i, model)
                logging_folder = os.path.join(temp_folder, model, str(i))
                local_config.setdefault("logging_folder", logging_folder)
                m = make(model, **shallow_copy_dict(local_config))
                local_pipelines.append(m.fit(x, y, x_cv, y_cv))
            pipelines_dict[model] = local_pipelines
    else:
        if num_jobs <= 1:
            print(
                f"{LoggingMixin.warning_prefix}we suggest setting `sequential` "
                f"to True when `num_jobs` is {num_jobs}"
            )
        # data
        data_folder = Experiment.dump_data_bundle(
            x,
            y,
            x_cv,
            y_cv,
            workplace=temp_folder,
        )
        # experiment
        experiment = Experiment(num_jobs=num_jobs)
        for model in models:
            for i in range(num_repeat):
                local_config = fetch_config(i, model)
                inject_distributed_tqdm_kwargs(i, num_jobs, local_config)
                experiment.add_task(
                    model=model,
                    compress=compress,
                    root_workplace=temp_folder,
                    config=local_config,
                    data_folder=data_folder,
                )
        # finalize
        results = experiment.run_tasks(use_tqdm=use_tqdm)
        if return_patterns:
            pipelines_dict = load_experiment_results(results)

    patterns = None
    if return_patterns:
        assert pipelines_dict is not None
        if predict_config is None:
            predict_config = {}
        patterns = {
            model: [m.to_pattern(**predict_config) for m in pipelines]
            for model, pipelines in pipelines_dict.items()
        }

    data = None
    if patterns is not None:
        data = patterns[models[0]][0].model.data

    return RepeatResult(data, experiment, pipelines_dict, patterns)


def make_toy_model(
    model: str = "fcnn",
    config: Optional[Dict[str, Any]] = None,
    *,
    task_type: task_type_type = "reg",
    data_tuple: Optional[Tuple[data_type, data_type]] = None,
    cuda: Optional[Union[int, str]] = "cpu",
) -> Pipeline:
    if config is None:
        config = {}
    if data_tuple is not None:
        x, y = data_tuple
        assert isinstance(x, list)
    else:
        if parse_task_type(task_type) is TaskTypes.REGRESSION:
            x, y = [[0]], [[1]]
        else:
            x, y = [[0], [1]], [[1], [0]]
    data_tuple = x, y
    model_config = {}
    if model in ("fcnn", "tree_dnn"):
        model_config = {
            "pipe_configs": {
                "fcnn": {
                    "head": {
                        "hidden_units": [100],
                        "mapping_configs": {"dropout": 0.0, "batch_norm": False},
                    }
                }
            }
        }
    base_config = {
        "model": model,
        "model_config": model_config,
        "cv_split": 0.0,
        "trigger_logging": False,
        "min_epoch": 1,
        "num_epoch": 2,
        "max_epoch": 4,
        "optimizer": "sgd",
        "optimizer_config": {"lr": 0.01},
        "task_type": task_type,
        "data_config": {
            "valid_columns": list(range(len(x[0]))),
            "label_process_method": "identical",
        },
        "verbose_level": 0,
        "cuda": cuda,
    }
    updated = update_dict(config, base_config)
    return make(**updated).fit(*data_tuple)


def switch_trainer_callback(callback_base: Type[TrainerCallback]) -> None:
    Trainer.callback_base = callback_base


__all__ = [
    "make",
    "make_external",
    "ModelConfig",
    "make_from",
    "finetune",
    "save",
    "load",
    "evaluate",
    "task_loader",
    "load_experiment_results",
    "inject_distributed_tqdm_kwargs",
    "repeat_with",
    "make_toy_model",
    "switch_trainer_callback",
    "Task",
    "Experiment",
    "ModelPattern",
    "EnsemblePattern",
    "RepeatResult",
]



import torch

from typing import Any
from typing import Dict
from typing import Optional

from ..base import HeadBase
from ....modules.blocks import DNDF
from ....modules.blocks import Linear


@HeadBase.register("dndf")
class DNDFHead(HeadBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        dndf_config: Optional[Dict[str, Any]],
        **kwargs: Any,
    ):
        super().__init__(in_dim, out_dim, **kwargs)
        self.dndf: Optional[DNDF]
        self.linear: Optional[torch.nn.Module]
        if dndf_config is not None:
            self.dndf = DNDF(in_dim, out_dim, **dndf_config)
            self.linear = None
        else:
            self.dndf = None
            self.linear = Linear(in_dim, out_dim)
            self.log_msg(  # type: ignore
                "`config` is not provided for `DNDFHead`, "
                "a `Linear` will be used instead of `DNDF`",
                self.warning_prefix,  # type: ignore
            )

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        if self.dndf is None:
            assert self.linear is not None
            return self.linear(net)
        return self.dndf(net)


__all__ = ["DNDFHead"]



import numpy as np

from typing import Any
from typing import Dict
from cftool.misc import shallow_copy_dict
from sklearn.tree import DecisionTreeClassifier

from ..base import HeadConfigs
from ....misc.toolkit import to_torch


@HeadConfigs.register("nnb_meta", "default")
class DefaultNNBMetaConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        # prepare
        x, y = self.tr_data.processed.xy
        split = self.dimensions.split_features(to_torch(x), np.arange(len(x)), "tr")
        if not self.dimensions.has_numerical:
            numerical = None
        else:
            numerical = split.numerical
        if self.dimensions.one_hot_dim == 0:
            categorical = None
        else:
            assert split.categorical is not None
            categorical = split.categorical.one_hot
        common_config = {"y_ravel": y.ravel()}
        # mnb
        mnb_config = shallow_copy_dict(common_config)
        mnb_config.update(
            {
                "categorical": categorical,
                "categorical_dims": self.dimensions.categorical_dims,
            }
        )
        # normal
        normal_config = shallow_copy_dict(common_config)
        normal_config.update({"numerical": numerical, "pretrain": True})
        return {"nnb_mnb": mnb_config, "nnb_normal": normal_config}


@HeadConfigs.register("ndt", "default")
class DefaultNDTConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        # decision tree
        x, y = self.tr_data.processed.xy
        msg = "fitting decision tree"
        self.log_msg(msg, self.info_prefix, verbose_level=2)  # type: ignore
        split = self.dimensions.split_features(to_torch(x), np.arange(len(x)), "tr")
        x_merge = split.merge().cpu().numpy()
        dt = DecisionTreeClassifier(max_depth=10, random_state=142857)
        dt.fit(x_merge, y.ravel(), sample_weight=self.tr_weights)
        return {
            "dt": dt,
            "activations": {"planes": "sign", "routes": "multiplied_softmax"},
            "activation_configs": {},
        }


__all__ = [
    "DefaultNNBMetaConfig",
    "DefaultNDTConfig",
]



import numpy as np

from typing import Any
from typing import Optional

from .base import ModelBase
from ..types import tensor_dict_type
from ..protocol import TrainerState


@ModelBase.register("fcnn")
@ModelBase.register_pipe("fcnn")
class FCNN(ModelBase):
    pass


@ModelBase.register("q_fcnn")
@ModelBase.register_pipe("fcnn")
class QuantileFCNN(ModelBase):
    def _init_config(self) -> None:
        super()._init_config()
        quantiles = self.config.setdefault("quantiles", [10, 30, 50, 70, 90])
        quantiles = list(map(int, map(round, quantiles)))
        try:
            self.median_idx = quantiles.index(50)
        except ValueError:
            raise ValueError("median (50) should be included in `quantiles`")
        self.quantiles = quantiles
        self.config["loss"] = "quantile"
        self.config["loss_config"] = {"q": [q / 100 for q in quantiles]}
        head_config = self.get_pipe_config("fcnn", "head")
        head_config["out_dim"] = len(quantiles)

    def forward(
        self,
        batch: tensor_dict_type,
        batch_idx: Optional[int] = None,
        state: Optional[TrainerState] = None,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        results = super().forward(
            batch,
            batch_idx,
            state,
            batch_indices,
            loader_name,
            **kwargs,
        )
        predictions = results["predictions"]
        median = predictions[..., [self.median_idx]]
        results = {"predictions": median, "quantiles": predictions}
        q = kwargs.get("q")
        if q is not None:
            fn = lambda q_: int(round(100 * q_))
            if not isinstance(q, (list, tuple)):
                q_list = [fn(q)]
            else:
                q_list = list(map(fn, q))
            indices = []
            for q in q_list:
                try:
                    indices.append(self.quantiles.index(q))
                except ValueError:
                    msg = f"quantile '{q}' is not included in the preset quantiles"
                    raise ValueError(msg)
            results["quantiles"] = predictions[..., indices]
        return results


__all__ = ["FCNN", "QuantileFCNN"]



import torch
import torch.nn as nn

from typing import *
from functools import partial
from torch.nn.functional import softplus

from ..types import tensor_dict_type


class BN(nn.BatchNorm1d):
    def forward(self, net: torch.Tensor) -> torch.Tensor:
        if len(net.shape) == 3:
            net = net.transpose(1, 2)
        net = super().forward(net)
        if len(net.shape) == 3:
            net = net.transpose(1, 2)
        return net


class Dropout(nn.Module):
    def __init__(self, dropout: float):
        if dropout < 0.0 or dropout > 1.0:
            msg = f"dropout probability has to be between 0 and 1, but got {dropout}"
            raise ValueError(msg)
        super().__init__()
        self._keep_prob = 1.0 - dropout
        self._mask_cache: Optional[torch.Tensor] = None

    def forward(self, net: torch.Tensor, *, reuse: bool = False) -> torch.Tensor:
        if not self.training:
            return net
        if reuse:
            mask = self._mask_cache
        else:
            self._mask_cache = mask = (
                torch.bernoulli(net.new(*net.shape).fill_(self._keep_prob))
                / self._keep_prob
            )
        net = net * mask
        del mask
        return net

    def extra_repr(self) -> str:
        return f"keep={self._keep_prob}"


class EMA(nn.Module):
    def __init__(
        self,
        decay: float,
        named_parameters: List[Tuple[str, nn.Parameter]],
    ):
        super().__init__()
        self._decay = decay
        self._named_parameters = named_parameters
        for name, param in self.tgt_params:
            self.register_buffer(self.get_name(True, name), param.data.clone())
            self.register_buffer(self.get_name(False, name), param.data.clone())

    @staticmethod
    def get_name(train: bool, name: str) -> str:
        prefix = "tr" if train else "ema"
        return f"{prefix}_{name}"

    @property
    def tgt_params(self) -> Iterator[Tuple[str, nn.Parameter]]:
        return map(
            lambda pair: (pair[0].replace(".", "_"), pair[1]),
            self._named_parameters,
        )

    def forward(self) -> None:
        for name, param in self.tgt_params:
            setattr(self, self.get_name(True, name), param.data.clone())
            ema_name = self.get_name(False, name)
            ema_attr = getattr(self, ema_name)
            ema = (1.0 - self._decay) * param.data + self._decay * ema_attr
            setattr(self, ema_name, ema.clone())

    def train(self, mode: bool = True) -> "EMA":
        super().train(mode)
        for name, param in self.tgt_params:
            param.data = getattr(self, self.get_name(mode, name)).clone()
        return self

    def extra_repr(self) -> str:
        max_str_len = max(len(name) for name, _ in self.tgt_params)
        return "\n".join(
            [f"(0): decay_rate={self._decay}\n(1): Params("]
            + [
                f"  {name:<{max_str_len}s} - torch.Tensor({list(param.shape)})"
                for name, param in self.tgt_params
            ]
            + [")"]
        )


class MTL(nn.Module):
    def __init__(
        self,
        num_tasks: int,
        method: Optional[str] = None,
    ):
        super().__init__()
        self._n_task, self._method = num_tasks, method
        if method is None or method == "naive":
            pass
        elif method == "softmax":
            self.w = torch.nn.Parameter(torch.ones(num_tasks))
        else:
            raise NotImplementedError(f"MTL method '{method}' not implemented")
        self.registered = False
        self._slice: Optional[int] = None
        self._registered: Dict[str, int] = {}

    def register(self, names: List[str]) -> None:
        if self.registered:
            raise ValueError("re-register is not permitted")
        self._rev_registered = {}
        for name in sorted(names):
            idx = len(self._registered)
            self._registered[name], self._rev_registered[idx] = idx, name
        self._slice, self.registered = len(names), True
        if self._slice > self._n_task:
            raise ValueError("registered names are more than n_task")

    def forward(
        self,
        loss_dict: tensor_dict_type,
        naive: bool = False,
    ) -> torch.Tensor:
        if not self.registered:
            raise ValueError("losses need to be registered")
        if naive or self._method is None:
            return self._naive(loss_dict)
        return getattr(self, f"_{self._method}")(loss_dict)

    @staticmethod
    def _naive(loss_dict: tensor_dict_type) -> torch.Tensor:
        return sum(loss_dict.values())  # type: ignore

    def _softmax(self, loss_dict: tensor_dict_type) -> torch.Tensor:
        assert self._slice is not None
        w = self.w if self._slice == self._n_task else self.w[: self._slice]
        softmax_w = nn.functional.softmax(w, dim=0)
        losses = []
        for key, loss in loss_dict.items():
            idx = self._registered.get(key)
            losses.append(loss if idx is None else loss * softmax_w[idx])
        final_loss: torch.Tensor = sum(losses)  # type: ignore
        return final_loss * self._slice

    def extra_repr(self) -> str:
        method = "naive" if self._method is None else self._method
        return f"n_task={self._n_task}, method='{method}'"


class Pruner(nn.Module):
    def __init__(self, config: Dict[str, Any], w_shape: Optional[List[int]] = None):
        super().__init__()
        self.eps: torch.Tensor
        self.exp: torch.Tensor
        self.alpha: Union[torch.Tensor, nn.Parameter]
        self.beta: Union[torch.Tensor, nn.Parameter]
        self.gamma: Union[torch.Tensor, nn.Parameter]
        self.max_ratio: Union[torch.Tensor, nn.Parameter]
        tensor = partial(torch.tensor, dtype=torch.float32)
        self.method = config.setdefault("method", "auto_prune")
        if self.method == "surgery":
            if w_shape is None:
                msg = "`w_shape` of `Pruner` should be provided when `surgery` is used"
                raise ValueError(msg)
            self.register_buffer("mask", torch.ones(*w_shape, dtype=torch.float32))
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 1.0)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 4.0)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1e-4)]))
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            keys = ["alpha", "beta", "gamma", "eps"]
        elif self.method == "simplified":
            self.register_buffer("alpha", tensor([config.setdefault("alpha", 0.01)]))
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.0)]))
            self.register_buffer(
                "max_ratio", tensor([config.setdefault("max_ratio", 1.0)])
            )
            self.register_buffer("exp", tensor([config.setdefault("exp", 0.5)]))
            keys = ["alpha", "beta", "max_ratio", "exp"]
        else:
            self.register_buffer(
                "alpha",
                tensor(
                    [
                        config.setdefault(
                            "alpha", 1e-4 if self.method == "hard_prune" else 1e-2
                        )
                    ]
                ),
            )
            self.register_buffer("beta", tensor([config.setdefault("beta", 1.0)]))
            self.register_buffer("gamma", tensor([config.setdefault("gamma", 1.0)]))
            self.register_buffer(
                "max_ratio", tensor([config.setdefault("max_ratio", 1.0)])
            )
            if not all(
                scalar > 0
                for scalar in [self.alpha, self.beta, self.gamma, self.max_ratio]
            ):
                raise ValueError("parameters should greater than 0. in pruner")
            self.register_buffer("eps", tensor([config.setdefault("eps", 1e-12)]))
            if self.method == "auto_prune":
                for attr in ["alpha", "beta", "gamma", "max_ratio"]:
                    setattr(self, attr, torch.log(torch.exp(getattr(self, attr)) - 1))
                self.alpha, self.beta, self.gamma, self.max_ratio = map(
                    lambda param: nn.Parameter(param),
                    [self.alpha, self.beta, self.gamma, self.max_ratio],
                )
            keys = ["alpha", "beta", "gamma", "max_ratio", "eps"]
        self._repr_keys = keys

    def forward(self, w: torch.Tensor) -> torch.Tensor:
        w_abs = torch.abs(w)
        if self.method == "surgery":
            mu, std = torch.mean(w_abs), torch.std(w_abs)
            zeros_mask = self.mask == 0.0
            ones_mask = self.mask == 1.0
            to_zeros_mask = ones_mask & (w_abs <= 0.9 * (mu - self.beta * std))
            to_ones_mask = zeros_mask & (w_abs >= 1.1 * (mu + self.beta * std))
            self.mask.masked_fill(to_zeros_mask, 0.0)  # type: ignore
            self.mask.masked_fill(to_ones_mask, 1.0)  # type: ignore
            mask = self.mask
            del mu, std, ones_mask, zeros_mask, to_zeros_mask, to_ones_mask
        else:
            if self.method != "auto_prune":
                alpha, beta, ratio = self.alpha, self.beta, self.max_ratio
            else:
                alpha, beta, ratio = map(
                    softplus, [self.alpha, self.beta, self.max_ratio]
                )
            if self.method == "simplified":
                log_w = torch.min(ratio, beta * torch.pow(w_abs, self.exp))
            else:
                w_abs_mean = torch.mean(w_abs)
                gamma = (
                    self.gamma if self.method != "auto_prune" else softplus(self.gamma)
                )
                log_w = torch.log(torch.max(self.eps, w_abs / (w_abs_mean * gamma)))
                log_w = torch.min(ratio, beta * log_w)
                del w_abs_mean
            mask = torch.max(alpha / beta * log_w, log_w)
            del log_w
        w = w * mask
        del w_abs, mask
        return w

    def extra_repr(self) -> str:
        if self.method == "auto_prune":
            return f"method='{self.method}'"
        max_str_len = max(map(len, self._repr_keys))
        return "\n".join(
            [f"(0): method={self.method}\n(1): Settings("]
            + [
                f"  {key:<{max_str_len}s} - {getattr(self, key).item()}"
                for key in self._repr_keys
            ]
            + [")"]
        )


__all__ = ["BN", "Dropout", "EMA", "MTL", "Pruner"]



import copy
import torch

import numpy as np

from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cfdata.types import np_int_type
from cfdata.types import np_float_type
from cfdata.tabular import DataLoader
from cfdata.tabular import ImbalancedSampler
from cfdata.tabular import TabularData as TD

from ..types import loader_batch_type
from ..protocol import DataProtocol
from ..protocol import SamplerProtocol
from ..protocol import DataLoaderProtocol
from ..misc.toolkit import to_torch


@DataProtocol.register("tabular")
class TabularData(TD, DataProtocol):
    pass


@SamplerProtocol.register("tabular")
class TabularSampler(ImbalancedSampler, SamplerProtocol):
    pass


@DataLoaderProtocol.register("tabular")
class TabularLoader(DataLoader, DataLoaderProtocol):
    def __next__(self) -> loader_batch_type:
        sample = DataLoader.__next__(self)
        if self.return_indices:
            (x_batch, labels), indices = sample
        else:
            x_batch, labels = sample
            indices = None
        x_batch = x_batch.astype(np_float_type)
        if self.is_onnx:
            if labels is None:
                labels = np.zeros([*x_batch.shape[:-1], 1], np_int_type)
            arrays = [x_batch, labels]
        else:
            x_batch = to_torch(x_batch)
            if labels is not None:
                labels = to_torch(labels)
                if self.data.is_clf:
                    labels = labels.to(torch.long)
            arrays = [x_batch, labels]

        sample = dict(zip(["x_batch", self.labels_key], arrays))
        if not self.return_indices:
            return sample
        assert indices is not None
        return sample, indices

    def copy(self) -> "DataLoader":
        copied_tabular_loader = copy.copy(self)
        copied_loader = super().copy()
        shallow_copied = shallow_copy_dict(copied_loader.__dict__)
        update_dict(shallow_copied, copied_tabular_loader.__dict__)
        return copied_tabular_loader


__all__ = [
    "TabularData",
    "TabularSampler",
    "TabularLoader",
]



from .base import ModelBase


@ModelBase.register("rnn")
@ModelBase.register_pipe("rnn", head="fcnn")
class RNN(ModelBase):
    pass


@ModelBase.register("tree_rnn")
@ModelBase.register_pipe("rnn_dndf", extractor="rnn", head="dndf")
@ModelBase.register_pipe("rnn_fcnn", extractor="rnn", head="fcnn", head_config="pruned")
class TreeRNN(ModelBase):
    pass


__all__ = ["RNN"]
